from __future__ import unicode_literals

from djblets.staticbundles import (
    PIPELINE_JAVASCRIPT as DJBLETS_PIPELINE_JAVASCRIPT,
    PIPELINE_STYLESHEETS as DJBLETS_PIPELINE_STYLESHEETS)


# Media compression
PIPELINE_JAVASCRIPT = dict({
    '3rdparty': {
        'source_filenames': (
            'lib/js/babel-polyfill-6.5.0.js',
            'lib/js/flot/jquery.flot.min.js',
            'lib/js/flot/jquery.flot.pie.min.js',
            'lib/js/flot/jquery.flot.selection.min.js',
            'lib/js/flot/jquery.flot.time.min.js',
            'lib/js/underscore-1.8.3.min.js',
            'lib/js/backbone-1.0.0.min.js',
            'lib/js/jquery.cookie-1.4.1.js',
            'lib/js/jquery.form.js',
            'lib/js/jquery.timesince.js',
            'lib/js/moment-2.12.0.js',
            'lib/js/moment-timezone-0.5.2.js',
            'lib/js/selectize-0.12.1.js',
            'lib/js/ui.autocomplete.js',
            'lib/js/codemirror-5.26.min.js',
        ),
        'output_filename': 'lib/js/3rdparty.min.js',
    },
    '3rdparty-jsonlint': {
        'source_filenames': (
            'lib/js/jsonlint.js',
        ),
        'output_filename': 'lib/js/3rdparty-jsonlint.min.js',
    },
    'js-test-libs': {
        'source_filenames': (
            'lib/js/jasmine-2.4.1.js',
            'lib/js/jasmine-html-2.4.1.js',
            'lib/js/jasmine-boot-2.4.1.js',
            'lib/js/jasmine.suites-1.0.js',
            'lib/js/jasmine.hide-filtered-1.0.js',
            'lib/js/jasmine.sourcemaps-1.0.js',
            'lib/js/sourcemapped-stacktrace.js',
        ),
        'output_filename': 'rb/js/js-test-libs.min.js',
    },
    'js-tests': {
        'source_filenames': (
            'rb/js/collections/tests/filteredCollectionTests.js',
            'rb/js/configForms/models/tests/resourceListItemModelTests.js',
            'rb/js/diffviewer/collections/tests/diffReviewableCollectionTests.es6.js',
            'rb/js/diffviewer/models/tests/diffFileModelTests.js',
            'rb/js/diffviewer/models/tests/diffReviewableModelTests.js',
            'rb/js/diffviewer/models/tests/diffRevisionModelTests.js',
            'rb/js/diffviewer/models/tests/paginationModelTests.js',
            'rb/js/diffviewer/views/tests/diffReviewableViewTests.js',
            'rb/js/models/tests/commentEditorModelTests.js',
            'rb/js/models/tests/extraDataTests.js',
            'rb/js/models/tests/reviewRequestEditorModelTests.js',
            'rb/js/models/tests/uploadDiffModelTests.js',
            'rb/js/models/tests/userSessionModelTests.js',
            'rb/js/newReviewRequest/views/tests/branchesViewTests.js',
            'rb/js/newReviewRequest/views/tests/postCommitViewTests.js',
            'rb/js/newReviewRequest/views/tests/repositorySelectionViewTests.js',
            'rb/js/pages/models/tests/diffViewerPageModelTests.es6.js',
            'rb/js/pages/models/tests/pageManagerModelTests.js',
            'rb/js/pages/models/tests/reviewablePageModelTests.es6.js',
            'rb/js/pages/views/tests/diffViewerPageViewTests.es6.js',
            'rb/js/pages/views/tests/reviewablePageViewTests.es6.js',
            'rb/js/resources/collections/tests/repositoryBranchesCollectionTests.js',
            'rb/js/resources/collections/tests/repositoryCommitsCollectionTests.js',
            'rb/js/resources/collections/tests/resourceCollectionTests.js',
            'rb/js/resources/models/tests/baseCommentModelTests.js',
            'rb/js/resources/models/tests/baseCommentReplyModelTests.js',
            'rb/js/resources/models/tests/baseResourceModelTests.js',
            'rb/js/resources/models/tests/defaultReviewerModelTests.js',
            'rb/js/resources/models/tests/diffCommentModelTests.js',
            'rb/js/resources/models/tests/draftReviewModelTests.js',
            'rb/js/resources/models/tests/draftReviewRequestModelTests.js',
            'rb/js/resources/models/tests/fileAttachmentModelTests.js',
            'rb/js/resources/models/tests/fileAttachmentCommentModelTests.js',
            'rb/js/resources/models/tests/fileDiffModelTests.js',
            'rb/js/resources/models/tests/generalCommentModelTests.js',
            'rb/js/resources/models/tests/screenshotModelTests.js',
            'rb/js/resources/models/tests/screenshotCommentModelTests.js',
            'rb/js/resources/models/tests/repositoryBranchModelTests.js',
            'rb/js/resources/models/tests/repositoryCommitModelTests.js',
            'rb/js/resources/models/tests/reviewGroupModelTests.js',
            'rb/js/resources/models/tests/reviewModelTests.js',
            'rb/js/resources/models/tests/reviewReplyModelTests.js',
            'rb/js/resources/models/tests/reviewRequestModelTests.js',
            'rb/js/resources/models/tests/userFileAttachmentModelTests.js',
            'rb/js/resources/models/tests/validateDiffModelTests.js',
            'rb/js/reviewRequestPage/models/tests/changeEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/entryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/reviewEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/reviewReplyEditorModelTests.js',
            'rb/js/reviewRequestPage/models/tests/reviewRequestPageModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/statusUpdatesEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/baseStatusUpdatesEntryViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/issueSummaryTableViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewEntryViewTests.js',
            'rb/js/reviewRequestPage/views/tests/reviewReplyDraftBannerViewTests.js',
            'rb/js/reviewRequestPage/views/tests/reviewReplyEditorViewTests.js',
            'rb/js/reviewRequestPage/views/tests/reviewRequestPageViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewViewTests.es6.js',
            'rb/js/ui/views/tests/dialogViewTests.js',
            'rb/js/ui/views/tests/infoboxManagerViewTests.es6.js',
            'rb/js/ui/views/tests/notificationManagerTests.es6.js',
            'rb/js/ui/views/tests/scrollManagerViewTests.es6.js',
            'rb/js/ui/views/tests/textEditorViewTests.js',
            'rb/js/utils/tests/dataUtilsTests.es6.js',
            'rb/js/utils/tests/keyBindingUtilsTests.js',
            'rb/js/utils/tests/linkifyUtilsTests.es6.js',
            'rb/js/utils/tests/urlUtilsTests.js',
            'rb/js/views/tests/collectionViewTests.js',
            'rb/js/views/tests/commentDialogViewTests.js',
            'rb/js/views/tests/commentIssueBarViewTests.js',
            'rb/js/views/tests/diffFragmentQueueViewTests.es6.js',
            'rb/js/views/tests/diffFragmentViewTests.es6.js',
            'rb/js/views/tests/draftReviewBannerViewTests.js',
            'rb/js/views/tests/fileAttachmentThumbnailViewTests.js',
            'rb/js/views/tests/reviewDialogViewTests.js',
            'rb/js/views/tests/reviewRequestEditorViewTests.js',
            'rb/js/views/tests/reviewRequestFieldViewsTests.es6.js',
            'rb/js/views/tests/screenshotThumbnailViewTests.js',
        ),
        'output_filename': 'rb/js/js-tests.min.js',
    },
    'common': {
        'source_filenames': (
            'rb/js/utils/consoleUtils.es6.js',
            'rb/js/utils/underscoreUtils.es6.js',
            'rb/js/common.es6.js',
            'rb/js/utils/apiErrors.js',
            'rb/js/utils/apiUtils.es6.js',
            'rb/js/utils/dataUtils.es6.js',
            'rb/js/utils/linkifyUtils.es6.js',
            'rb/js/utils/mathUtils.es6.js',
            'rb/js/utils/keyBindingUtils.es6.js',
            'rb/js/utils/urlUtils.js',
            'rb/js/collections/baseCollection.es6.js',
            'rb/js/collections/filteredCollection.es6.js',
            'rb/js/extensions/models/aliases.js',
            'rb/js/extensions/models/commentDialogHookModel.js',
            'rb/js/extensions/models/reviewDialogCommentHookModel.js',
            'rb/js/extensions/models/reviewDialogHookModel.js',
            'rb/js/extensions/models/reviewRequestActionHookModel.js',
            'rb/js/pages/models/pageManagerModel.js',
            'rb/js/pages/models/pageModel.es6.js',
            'rb/js/pages/views/basePageView.es6.js',
            'rb/js/models/extraDataModel.js',
            'rb/js/models/extraDataMixin.js',
            'rb/js/resources/utils/serializers.es6.js',
            'rb/js/resources/models/baseResourceModel.es6.js',
            'rb/js/resources/models/apiTokenModel.es6.js',
            'rb/js/resources/models/repositoryBranchModel.es6.js',
            'rb/js/resources/models/repositoryCommitModel.es6.js',
            'rb/js/resources/models/draftResourceChildModelMixin.es6.js',
            'rb/js/resources/models/draftResourceModelMixin.es6.js',
            'rb/js/resources/models/draftReviewRequestModel.es6.js',
            'rb/js/resources/models/reviewModel.es6.js',
            'rb/js/resources/models/draftReviewModel.es6.js',
            'rb/js/resources/models/baseCommentModel.es6.js',
            'rb/js/resources/models/baseCommentReplyModel.es6.js',
            'rb/js/resources/models/defaultReviewerModel.es6.js',
            'rb/js/resources/models/diffCommentModel.es6.js',
            'rb/js/resources/models/diffCommentReplyModel.es6.js',
            'rb/js/resources/models/diffModel.es6.js',
            'rb/js/resources/models/fileAttachmentModel.es6.js',
            'rb/js/resources/models/fileAttachmentCommentModel.es6.js',
            'rb/js/resources/models/fileAttachmentCommentReplyModel.es6.js',
            'rb/js/resources/models/generalCommentModel.es6.js',
            'rb/js/resources/models/generalCommentReplyModel.es6.js',
            'rb/js/resources/models/fileDiffModel.es6.js',
            'rb/js/resources/models/draftFileAttachmentModel.es6.js',
            'rb/js/resources/models/repositoryModel.es6.js',
            'rb/js/resources/models/reviewGroupModel.es6.js',
            'rb/js/resources/models/reviewReplyModel.es6.js',
            'rb/js/resources/models/reviewRequestModel.es6.js',
            'rb/js/resources/models/screenshotModel.es6.js',
            'rb/js/resources/models/screenshotCommentModel.es6.js',
            'rb/js/resources/models/screenshotCommentReplyModel.es6.js',
            'rb/js/resources/models/userFileAttachmentModel.es6.js',
            'rb/js/resources/models/validateDiffModel.es6.js',
            'rb/js/resources/collections/resourceCollection.es6.js',
            'rb/js/resources/collections/repositoryBranchesCollection.es6.js',
            'rb/js/resources/collections/repositoryCommitsCollection.es6.js',
            'rb/js/ui/views/centeredElementManager.es6.js',
            'rb/js/ui/views/dialogView.es6.js',
            'rb/js/ui/views/baseInfoboxView.es6.js',
            'rb/js/ui/views/infoboxManagerView.es6.js',
            'rb/js/ui/views/bugInfoboxView.es6.js',
            'rb/js/ui/views/inlineEditorView.es6.js',
            'rb/js/ui/views/notificationManager.es6.js',
            'rb/js/ui/views/reviewRequestInfoboxView.es6.js',
            'rb/js/ui/views/scrollManagerView.es6.js',
            'rb/js/ui/views/splitButtonView.es6.js',
            'rb/js/ui/views/textEditorView.es6.js',
            'rb/js/ui/views/userInfoboxView.es6.js',
            'rb/js/models/starManagerModel.js',
            'rb/js/models/userSessionModel.js',
            'rb/js/views/headerView.js',
            'rb/js/views/collectionView.es6.js',
            'rb/js/views/starManagerView.js',
        ),
        'output_filename': 'rb/js/base.min.js',
    },
    'account-page': {
        'source_filenames': (
            'rb/js/accountPrefsPage/views/apiTokensView.es6.js',
            'rb/js/accountPrefsPage/views/joinedGroupsView.es6.js',
            'rb/js/accountPrefsPage/views/oauthApplicationsView.es6.js',
            'rb/js/accountPrefsPage/views/oauthTokensView.es6.js',
        ),
        'output_filename': 'rb/js/account-page.min.js',
    },
    'config-forms': {
        'source_filenames': (
            'rb/js/configForms/base.js',
            'rb/js/configForms/models/resourceListItemModel.es6.js',
        ),
        'output_filename': 'rb/js/config-forms.min.js',
    },
    'datagrid-pages': {
        'source_filenames': (
            'rb/js/pages/models/datagridPageModel.js',
            'rb/js/pages/models/dashboardModel.js',
            'rb/js/pages/views/datagridPageView.js',
            'rb/js/pages/views/dashboardView.js',
        ),
        'output_filename': 'rb/js/dashboard.min.js',
    },
    'reviews': {
        'source_filenames': (
            # Note: These are roughly in dependency order.
            'rb/js/models/abstractCommentBlockModel.es6.js',
            'rb/js/models/abstractReviewableModel.es6.js',
            'rb/js/models/commentEditorModel.es6.js',
            'rb/js/models/commentIssueManagerModel.es6.js',
            'rb/js/models/fileAttachmentCommentBlockModel.es6.js',
            'rb/js/models/fileAttachmentReviewableModel.es6.js',
            'rb/js/models/regionCommentBlockModel.es6.js',
            'rb/js/models/reviewRequestEditorModel.es6.js',
            'rb/js/models/imageReviewableModel.es6.js',
            'rb/js/models/dummyReviewableModel.es6.js',
            'rb/js/models/screenshotCommentBlockModel.es6.js',
            'rb/js/models/screenshotReviewableModel.es6.js',
            'rb/js/models/textBasedCommentBlockModel.es6.js',
            'rb/js/models/textBasedReviewableModel.es6.js',
            'rb/js/models/uploadDiffModel.js',
            'rb/js/pages/models/reviewablePageModel.es6.js',
            'rb/js/pages/models/diffViewerPageModel.es6.js',
            'rb/js/pages/views/reviewablePageView.es6.js',
            'rb/js/pages/views/diffViewerPageView.es6.js',
            'rb/js/utils/textUtils.es6.js',
            'rb/js/views/abstractCommentBlockView.es6.js',
            'rb/js/views/abstractReviewableView.es6.js',
            'rb/js/views/commentDialogView.es6.js',
            'rb/js/views/commentIssueBarView.es6.js',
            'rb/js/views/diffFragmentQueueView.es6.js',
            'rb/js/views/diffFragmentView.es6.js',
            'rb/js/views/dndUploaderView.es6.js',
            'rb/js/views/draftReviewBannerView.es6.js',
            'rb/js/views/uploadAttachmentView.js',
            'rb/js/views/revisionSelectorView.js',
            'rb/js/views/fileAttachmentReviewableView.es6.js',
            'rb/js/views/fileAttachmentRevisionLabelView.es6.js',
            'rb/js/views/fileAttachmentRevisionSelectorView.js',
            'rb/js/views/fileAttachmentThumbnailView.js',
            'rb/js/views/floatingBannerView.js',
            'rb/js/views/regionCommentBlockView.es6.js',
            'rb/js/views/reviewDialogView.es6.js',
            'rb/js/views/reviewRequestFieldViews.es6.js',
            'rb/js/views/reviewRequestEditorView.es6.js',
            'rb/js/views/screenshotThumbnailView.js',
            'rb/js/views/imageReviewableView.es6.js',
            'rb/js/views/dummyReviewableView.es6.js',
            'rb/js/views/textBasedCommentBlockView.es6.js',
            'rb/js/views/textBasedReviewableView.es6.js',
            'rb/js/views/textCommentRowSelector.js',
            'rb/js/views/markdownReviewableView.es6.js',
            'rb/js/views/uploadDiffView.js',
            'rb/js/views/updateDiffView.js',
            'rb/js/diffviewer/models/diffCommentBlockModel.es6.js',
            'rb/js/diffviewer/models/diffCommentsHintModel.js',
            'rb/js/diffviewer/models/diffFileModel.js',
            'rb/js/diffviewer/models/diffReviewableModel.es6.js',
            'rb/js/diffviewer/models/diffRevisionModel.js',
            'rb/js/diffviewer/models/paginationModel.js',
            'rb/js/diffviewer/collections/diffFileCollection.js',
            'rb/js/diffviewer/collections/diffReviewableCollection.es6.js',
            'rb/js/diffviewer/views/chunkHighlighterView.js',
            'rb/js/diffviewer/views/diffCommentBlockView.es6.js',
            'rb/js/diffviewer/views/diffCommentsHintView.js',
            'rb/js/diffviewer/views/diffComplexityIconView.js',
            'rb/js/diffviewer/views/diffFileIndexView.js',
            'rb/js/diffviewer/views/diffReviewableView.es6.js',
            'rb/js/diffviewer/views/diffRevisionLabelView.js',
            'rb/js/diffviewer/views/diffRevisionSelectorView.js',
            'rb/js/diffviewer/views/paginationView.js',
        ),
        'output_filename': 'rb/js/reviews.min.js',
    },
    'review-request-page': {
        'source_filenames': (
            'rb/js/reviewRequestPage/base.js',
            'rb/js/reviewRequestPage/models/entryModel.es6.js',
            'rb/js/reviewRequestPage/models/reviewEntryModel.es6.js',
            'rb/js/reviewRequestPage/models/reviewReplyEditorModel.js',
            'rb/js/reviewRequestPage/models/reviewRequestPageModel.es6.js',
            'rb/js/reviewRequestPage/models/statusUpdatesEntryModel.es6.js',
            'rb/js/reviewRequestPage/models/changeEntryModel.es6.js',
            'rb/js/reviewRequestPage/views/entryView.es6.js',
            'rb/js/reviewRequestPage/views/baseStatusUpdatesEntryView.es6.js',
            'rb/js/reviewRequestPage/views/changeEntryView.es6.js',
            'rb/js/reviewRequestPage/views/initialStatusUpdatesEntryView.es6.js',
            'rb/js/reviewRequestPage/views/issueSummaryTableView.es6.js',
            'rb/js/reviewRequestPage/views/reviewEntryView.es6.js',
            'rb/js/reviewRequestPage/views/reviewReplyDraftBannerView.js',
            'rb/js/reviewRequestPage/views/reviewReplyEditorView.js',
            'rb/js/reviewRequestPage/views/reviewRequestPageView.es6.js',
            'rb/js/reviewRequestPage/views/reviewView.es6.js',
        ),
        'output_filename': 'rb/js/review-request-page.min.js',
    },
    'newReviewRequest': {
        'source_filenames': (
            # Note: These are roughly in dependency order.
            'rb/js/models/uploadDiffModel.js',
            'rb/js/newReviewRequest/models/postCommitModel.js',
            'rb/js/newReviewRequest/models/newReviewRequestModel.js',
            'rb/js/views/uploadDiffView.js',
            'rb/js/newReviewRequest/views/branchView.js',
            'rb/js/newReviewRequest/views/branchesView.js',
            'rb/js/newReviewRequest/views/commitView.js',
            'rb/js/newReviewRequest/views/commitsView.js',
            'rb/js/newReviewRequest/views/repositoryView.js',
            'rb/js/newReviewRequest/views/repositorySelectionView.js',
            'rb/js/newReviewRequest/views/postCommitView.js',
            'rb/js/newReviewRequest/views/preCommitView.js',
            'rb/js/newReviewRequest/views/newReviewRequestView.js',
        ),
        'output_filename': 'rb/js/newReviewRequest.min.js',
    },
    'oauth-edit': {
        'source_filenames': (
            'rb/js/accountPrefsPage/views/oauthClientSecretView.es6.js',
        ),
        'output_filename': 'rb/js/oauth.min.js',
    },
    'admin-dashboard': {
        'source_filenames': (
            'lib/js/jquery.masonry.js',
            'rb/js/admin/admin.js',
            'rb/js/admin/views/supportBannerView.js',
        ),
        'output_filename': 'rb/js/admin-dashboard.min.js',
    },
    'repositoryform': {
        'source_filenames': (
            'rb/js/admin/repositoryform.js',
        ),
        'output_filename': 'rb/js/repositoryform.min.js',
    },
    'webhooks-form': {
        'source_filenames': (
            'rb/js/admin/views/webhookFormView.js',
        ),
        'output_filename': 'rb/js/webhooks-form.min.js',
    },
    'widgets': {
        'source_filenames': (
            'rb/js/admin/views/relatedObjectSelectorView.es6.js',
            'rb/js/admin/views/relatedUserSelectorView.es6.js',
        ),
        'output_filename': 'rb/js/widgets.min.js',
    },
}, **DJBLETS_PIPELINE_JAVASCRIPT)


PIPELINE_STYLESHEETS = dict({
    'common': {
        'source_filenames': (
            'lib/css/codemirror.css',
            'lib/css/jquery-ui-1.8.24.min.css',
            'lib/css/fontawesome.less',
            'lib/css/selectize.default-0.12.1.css',
            'rb/css/assets/icons.less',
            'rb/css/layout/helpers.less',
            'rb/css/pages/base.less',
            'rb/css/pages/search.less',
            'rb/css/ui/banners.less',
            'rb/css/ui/boxes.less',
            'rb/css/ui/buttons.less',
            'rb/css/ui/datagrids.less',
            'rb/css/ui/forms.less',
            'rb/css/ui/infoboxes.less',
            'rb/css/ui/markdown.less',
            'rb/css/ui/menus.less',
            'rb/css/ui/sidebars.less',
            'rb/css/common.less',
        ),
        'output_filename': 'rb/css/common.min.css',
        'absolute_paths': False,
    },
    'js-tests': {
        'source_filenames': (
            'lib/css/jasmine-2.4.1.css',
            'rb/css/pages/js-tests.less',
        ),
        'output_filename': 'rb/css/js-tests.min.css',
        'absolute_paths': False,
    },
    'account-page': {
        'source_filenames': (
            'rb/css/pages/my-account.less',
        ),
        'output_filename': 'rb/css/account-page.min.css',
    },
    'reviews': {
        'source_filenames': (
            'rb/css/pages/diffviewer.less',
            'rb/css/pages/image-review-ui.less',
            'rb/css/pages/text-review-ui.less',
            'rb/css/pages/reviews.less',
            'rb/css/pages/review-request.less',
            'rb/css/ui/dnd-uploader.less',
            'rb/css/syntax.less',
        ),
        'output_filename': 'rb/css/reviews.min.css',
        'absolute_paths': False,
    },
    'newReviewRequest': {
        'source_filenames': (
            'rb/css/pages/newReviewRequest.less',
        ),
        'output_filename': 'rb/css/newReviewRequest.min.css',
        'absolute_paths': False,
    },
    'oauth': {
        'source_filenames': (
            'rb/css/pages/oauth.less',
        ),
        'output_filename': 'rb/css/oauth.min.css',
        'absolute_paths': False,
    },
    'admin': {
        'source_filenames': (
            'rb/css/pages/admin.less',
            'rb/css/pages/admin-dashboard.less',
            'rb/css/pages/admin-integrations.less',
        ),
        'output_filename': 'rb/css/admin.min.css',
        'absolute_paths': False,
    },
    'widgets': {
        'source_filenames': (
            'rb/css/ui/related-object-selector.less',
        ),
        'output_filename': 'rb/css/widgets.min.css',
        'absolute_paths': False,
    },
}, **DJBLETS_PIPELINE_STYLESHEETS)



from __future__ import unicode_literals

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import CounterField, JSONField
from djblets.db.managers import ConcurrencyManager


@python_2_unicode_compatible
class BaseComment(models.Model):
    """The base class for all comment types."""

    OPEN = 'O'
    RESOLVED = 'R'
    DROPPED = 'D'
    VERIFYING_RESOLVED = 'A'
    VERIFYING_DROPPED = 'B'

    ISSUE_STATUSES = (
        (OPEN, _('Open')),
        (RESOLVED, _('Resolved')),
        (DROPPED, _('Dropped')),
        (VERIFYING_RESOLVED, _('Waiting for verification to resolve')),
        (VERIFYING_DROPPED, _('Waiting for verification to drop')),
    )

    ISSUE_STATUS_TO_STRING = {
        OPEN: 'open',
        RESOLVED: 'resolved',
        DROPPED: 'dropped',
        VERIFYING_RESOLVED: 'verifying-resolved',
        VERIFYING_DROPPED: 'verifying-dropped',
    }

    ISSUE_STRING_TO_STATUS = {
        'open': OPEN,
        'resolved': RESOLVED,
        'dropped': DROPPED,
        'verifying-resolved': VERIFYING_RESOLVED,
        'verifying-dropped': VERIFYING_DROPPED,
    }

    issue_opened = models.BooleanField(_('Issue Opened'), default=False)
    issue_status = models.CharField(_('Issue Status'),
                                    max_length=1,
                                    choices=ISSUE_STATUSES,
                                    blank=True,
                                    null=True,
                                    db_index=True)

    reply_to = models.ForeignKey('self', blank=True, null=True,
                                 related_name='replies',
                                 verbose_name=_('Reply To'))
    timestamp = models.DateTimeField(_('Timestamp'), default=timezone.now)
    text = models.TextField(_('Comment Text'))
    rich_text = models.BooleanField(_('Rich Text'), default=False)

    extra_data = JSONField(null=True)

    # Set this up with a ConcurrencyManager to help prevent race conditions.
    objects = ConcurrencyManager()

    @staticmethod
    def issue_status_to_string(status):
        """Return a string representation of the status field.

        Args:
            status (unicode):
                The value of the ``issue_status`` field.

        Returns:
            unicode:
            A string representation of the status used for the API and other
            interfaces.
        """
        try:
            return BaseComment.ISSUE_STATUS_TO_STRING[status]
        except KeyError:
            return ''

    @staticmethod
    def issue_string_to_status(status):
        """Return a DB representation of the given status string.

        Args:
            status (unicode):
                The status string to convert.

        Returns:
            unicode:
            A value suitable for storing in the ``issue_status`` field.
        """
        try:
            return BaseComment.ISSUE_STRING_TO_STATUS[status]
        except KeyError:
            raise Exception('Invalid issue status "%s"' % status)

    def _get_require_verification(self):
        return self.extra_data.get('require_verification', False)

    def _set_require_verification(self, value):
        if not isinstance(value, bool):
            raise ValueError('require_verification must be a bool')

        self.extra_data['require_verification'] = value

    require_verification = property(
        _get_require_verification, _set_require_verification,
        doc='Whether this comment requires verification before closing.')

    def __init__(self, *args, **kwargs):
        """Initialize the comment.

        Args:
            *args (tuple):
                Positional arguments to pass through to the model
                initialization.

            **kwargs (dict):
                Keyword arguments to pass through to the model
                initialization.
        """
        super(BaseComment, self).__init__(*args, **kwargs)

        self._loaded_issue_status = self.issue_status

    def get_review_request(self):
        """Return this comment's review request.

        Returns:
            reviewboard.reviews.models.review_request.ReviewRequest:
            The review request that this comment was made on.
        """
        if hasattr(self, '_review_request'):
            return self._review_request
        else:
            return self.get_review().review_request

    def get_review(self):
        """Return this comment's review.

        Returns:
            reviewboard.reviews.models.review.Review:
            The review containing this comment.
        """
        if hasattr(self, '_review'):
            return self._review
        else:
            return self.review.get()

    def get_review_url(self):
        """Return the URL to view this comment.

        Returns:
            unicode:
            The absolute URL to view this comment in the web UI.
        """
        return '%s#%s%d' % (self.get_review_request().get_absolute_url(),
                            self.anchor_prefix, self.id)

    def is_reply(self):
        """Return whether this comment is a reply to another comment.

        Returns:
            bool:
            True if the comment is a reply.
        """
        return self.reply_to_id is not None
    is_reply.boolean = True

    def is_accessible_by(self, user):
        """Return whether the user can access this comment.

        Args:
            user (django.contrib.auth.models.User):
                The user being checked.

        Returns:
            bool:
            True if the given user can access this comment.
        """
        return self.get_review().is_accessible_by(user)

    def is_mutable_by(self, user):
        """Return whether the user can modify this comment.

        Args:
            user (django.contrib.auth.models.User):
                The user being checked.

        Returns:
            bool:
            True if the given user can modify this comment.
        """
        return self.get_review().is_mutable_by(user)

    def public_replies(self, user=None):
        """Return the public replies to this comment.

        Args:
            user (django.contrib.auth.models.User, optional):
                A user to filter by, if desired. If specified, only replies
                authored by this user will be returned.

        Returns:
            list of reviewboard.reviews.models.base_comment.BaseComment:
            The public replies to this comment.
        """
        if hasattr(self, '_replies'):
            return self._replies

        if user and user.is_authenticated():
            return self.replies.filter(Q(review__public=True) |
                                       Q(review__user=user))
        else:
            return self.replies.filter(review__public=True)

    def can_change_issue_status(self, user):
        """Return whether the user can change the issue status.

        Currently, this is allowed for:
        - The user who owns the review request.
        - The user who opened the issue (posted the comment).

        Args:
            user (django.contrib.auth.models.User):
                The user being checked.

        Returns:
            bool:
            True if the given user is allowed to change the issue status.
        """
        if not (user and user.is_authenticated()):
            return False

        return (self.get_review_request().is_mutable_by(user) or
                user == self.get_review().user)

    def can_verify_issue_status(self, user):
        """Return whether the user can verify the issue status.

        Currently this is allowed for:

        - The user who opened the issue.
        - Administrators.

        Args:
            user (django.contrib.auth.models.User):
                The user being checked.

        Returns:
            bool:
            True if the given user is allowed to verify the issue status.
        """
        if not (user and user.is_authenticated()):
            return False

        review = self.get_review()
        local_site = review.review_request.local_site

        return (user.is_superuser or
                user.pk == review.user_id or
                (local_site and local_site.is_mutable_by(user)))

    def save(self, **kwargs):
        """Save the comment.

        Args:
            **kwargs (dict):
                Keyword arguments passed to the method (unused).
        """
        from reviewboard.reviews.models.review_request import ReviewRequest

        self.timestamp = timezone.now()

        super(BaseComment, self).save()

        try:
            # Update the review timestamp, but only if it's a draft.
            # Otherwise, resolving an issue will change the timestamp of
            # the review.
            review = self.get_review()

            if not review.public:
                review.timestamp = self.timestamp
                review.save()
            else:
                if (not self.is_reply() and
                    self.issue_opened and
                    self._loaded_issue_status != self.issue_status):
                    # The user has toggled the issue status of this comment,
                    # so update the issue counts for the review request.
                    old_field = ReviewRequest.ISSUE_COUNTER_FIELDS[
                        self._loaded_issue_status]
                    new_field = ReviewRequest.ISSUE_COUNTER_FIELDS[
                        self.issue_status]

                    if old_field != new_field:
                        CounterField.increment_many(
                            self.get_review_request(),
                            {
                                old_field: -1,
                                new_field: 1,
                            })

                q = ReviewRequest.objects.filter(pk=review.review_request_id)
                q.update(last_review_activity_timestamp=self.timestamp)
        except ObjectDoesNotExist:
            pass

    def __str__(self):
        """Return a string representation of the comment.

        Returns:
            unicode:
            A string representation of the comment.
        """
        return self.text

    class Meta:
        abstract = True
        app_label = 'reviews'
        ordering = ['timestamp']



"""Tests for top level Review Board modules."""

from __future__ import unicode_literals

import os

from django.utils import six
from djblets.staticbundles import (
    PIPELINE_JAVASCRIPT as DJBLETS_PIPELINE_JAVASCRIPT,
    PIPELINE_STYLESHEETS as DJBLETS_PIPELINE_STYLESHEETS)

from reviewboard.staticbundles import PIPELINE_JAVASCRIPT, PIPELINE_STYLESHEETS
from reviewboard.testing import TestCase


class StaticBundlesTests(TestCase):
    """Tests the static bundles in reviewboard.staticbundles."""

    def _check_file_groups(self, groups, exclude):
        """Checks that all source files in the given groups exist.

        Args:
            groups (dict):
                The groups to check.

            exclude (list):
                List of group names to exclude from the check.
        """
        missing = set()

        for name, group in six.iteritems(groups):
            if name in exclude:
                continue

            for path in group['source_filenames']:
                static_path = os.path.join('reviewboard', 'static', path)

                if not os.path.exists(static_path):
                    missing.add(path)

        self.assertSetEqual(missing, set())

    def test_static_javascript_files(self):
        """Testing that all static javascript files exist"""
        self._check_file_groups(PIPELINE_JAVASCRIPT,
                                DJBLETS_PIPELINE_JAVASCRIPT.keys())

    def test_static_stylesheet_files(self):
        """Testing that all static stylesheet files exist"""
        self._check_file_groups(PIPELINE_STYLESHEETS,
                                DJBLETS_PIPELINE_STYLESHEETS.keys())



from __future__ import unicode_literals

from django.utils import six

from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.service import (register_hosting_service,
                                             unregister_hosting_service)
from reviewboard.scmtools.forms import RepositoryForm
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.site.models import LocalSite
from reviewboard.testing.hosting_services import (SelfHostedTestService,
                                                  TestService)
from reviewboard.testing.testcase import TestCase


class RepositoryFormTests(TestCase):
    """Unit tests for the repository form."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(RepositoryFormTests, self).setUp()

        register_hosting_service('test', TestService)
        register_hosting_service('self_hosted_test', SelfHostedTestService)

        self.git_tool_id = Tool.objects.get(name='Git').pk

    def tearDown(self):
        super(RepositoryFormTests, self).tearDown()

        unregister_hosting_service('self_hosted_test')
        unregister_hosting_service('test')

    def test_plain_repository(self):
        """Testing RepositoryForm with a plain repository"""
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'custom',
            'tool': self.git_tool_id,
            'path': '/path/to/test.git',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.hosting_account, None)
        self.assertEqual(repository.extra_data, {})

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_plain_repository_with_missing_fields(self):
        """Testing RepositoryForm with a plain repository with missing fields
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'custom',
            'tool': self.git_tool_id,
            'bug_tracker_type': 'none',
        })

        self.assertFalse(form.is_valid())
        self.assertIn('path', form.errors)

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account(self):
        """Testing RepositoryForm with a hosting service and new account"""
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': 'testuser',
            'test-hosting_account_password': 'testpass',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name, 'test')
        self.assertEqual(repository.hosting_account.local_site, None)
        self.assertEqual(repository.extra_data['repository_plan'], '')
        self.assertEqual(repository.path, 'http://example.com/testrepo/')
        self.assertEqual(repository.mirror_path, '')

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_auth_error(self):
        """Testing RepositoryForm with a hosting service and new account and
        authorization error
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': 'baduser',
            'test-hosting_account_password': 'testpass',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertIn('hosting_account', form.errors)
        self.assertEqual(form.errors['hosting_account'],
                         ['Unable to link the account: The username is '
                          'very very bad.'])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_2fa_code_required(self):
        """Testing RepositoryForm with a hosting service and new account and
        two-factor auth code required
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': '2fa-user',
            'test-hosting_account_password': 'testpass',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertIn('hosting_account', form.errors)
        self.assertEqual(form.errors['hosting_account'],
                         ['Enter your 2FA code.'])
        self.assertTrue(
            form.hosting_service_info['test']['needs_two_factor_auth_code'])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_2fa_code_provided(self):
        """Testing RepositoryForm with a hosting service and new account and
        two-factor auth code provided
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'test-hosting_account_username': '2fa-user',
            'test-hosting_account_password': 'testpass',
            'test-hosting_account_two_factor_auth_code': '123456',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)
        self.assertFalse(
            form.hosting_service_info['test']['needs_two_factor_auth_code'])

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_new_account_missing_fields(self):
        """Testing RepositoryForm with a hosting service and new account and
        missing fields
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

        self.assertIn('hosting_account_username', form.errors)
        self.assertIn('hosting_account_password', form.errors)

        # Make sure the auth form also contains the errors.
        auth_form = form.hosting_auth_forms.pop('test')
        self.assertIn('hosting_account_username', auth_form.errors)
        self.assertIn('hosting_account_password', auth_form.errors)

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_self_hosted_and_new_account(self):
        """Testing RepositoryForm with a self-hosted hosting service and new
        account
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': 'https://myserver.com',
            'self_hosted_test-hosting_account_username': 'testuser',
            'self_hosted_test-hosting_account_password': 'testpass',
            'test_repo_name': 'myrepo',
            'tool': self.git_tool_id,
            'bug_tracker_type': 'none',
        })
        form.validate_repository = False

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.hosting_account.hosting_url,
                         'https://myserver.com')
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name,
                         'self_hosted_test')
        self.assertEqual(repository.hosting_account.local_site, None)
        self.assertEqual(repository.extra_data['test_repo_name'], 'myrepo')
        self.assertEqual(repository.extra_data['hosting_url'],
                         'https://myserver.com')
        self.assertEqual(repository.path, 'https://myserver.com/myrepo/')
        self.assertEqual(repository.mirror_path, 'git@myserver.com:myrepo/')

        # Make sure none of the other auth forms are unhappy. That would be
        # an indicator that we're doing form processing and validation wrong.
        for auth_form in six.itervalues(form.hosting_auth_forms):
            self.assertEqual(auth_form.errors, {})

    def test_with_hosting_service_self_hosted_and_blank_url(self):
        """Testing RepositoryForm with a self-hosted hosting service and blank
        URL
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': '',
            'self_hosted_test-hosting_account_username': 'testuser',
            'self_hosted_test-hosting_account_password': 'testpass',
            'test_repo_name': 'myrepo',
            'tool': self.git_tool_id,
            'bug_tracker_type': 'none',
        })
        form.validate_repository = False

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

    def test_with_hosting_service_new_account_localsite(self):
        """Testing RepositoryForm with a hosting service, new account and
        LocalSite
        """
        local_site = LocalSite.objects.create(name='testsite')

        form = RepositoryForm(
            {
                'name': 'test',
                'hosting_type': 'test',
                'test-hosting_account_username': 'testuser',
                'test-hosting_account_password': 'testpass',
                'tool': self.git_tool_id,
                'test_repo_name': 'testrepo',
                'bug_tracker_type': 'none',
                'local_site': local_site.pk,
            },
            local_site_name=local_site.name)

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.local_site, local_site)
        self.assertEqual(repository.hosting_account.username, 'testuser')
        self.assertEqual(repository.hosting_account.service_name, 'test')
        self.assertEqual(repository.hosting_account.local_site, local_site)
        self.assertEqual(repository.extra_data['repository_plan'], '')

    def test_with_hosting_service_existing_account(self):
        """Testing RepositoryForm with a hosting service and existing
        account
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(repository.extra_data['repository_plan'], '')

    def test_with_hosting_service_existing_account_needs_reauth(self):
        """Testing RepositoryForm with a hosting service and existing
        account needing re-authorization
        """
        # We won't be setting the password, so that is_authorized() will
        # fail.
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)
        self.assertEqual(set(form.errors.keys()),
                         set(['hosting_account_username',
                              'hosting_account_password']))

    def test_with_hosting_service_existing_account_reauthing(self):
        """Testing RepositoryForm with a hosting service and existing
        account with re-authorizating
        """
        # We won't be setting the password, so that is_authorized() will
        # fail.
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'test-hosting_account_username': 'testuser2',
            'test-hosting_account_password': 'testpass2',
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())
        self.assertTrue(form.hosting_account_linked)

        account = HostingServiceAccount.objects.get(pk=account.pk)
        self.assertEqual(account.username, 'testuser2')
        self.assertEqual(account.data['password'], 'testpass2')

    def test_with_hosting_service_self_hosted_and_existing_account(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        existing account
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'self_hosted_test-hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'myrepo',
            'bug_tracker_type': 'none',
        })
        form.validate_repository = False

        self.assertTrue(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

        repository = form.save()
        self.assertEqual(repository.name, 'test')
        self.assertEqual(repository.hosting_account, account)
        self.assertEqual(repository.extra_data['hosting_url'],
                         'https://example.com')

    def test_with_self_hosted_and_invalid_account_service(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        invalid existing account due to mismatched service type
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example1.com')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'myrepo',
            'bug_tracker_type': 'none',
        })
        form.validate_repository = False

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

    def test_with_self_hosted_and_invalid_account_local_site(self):
        """Testing RepositoryForm with a self-hosted hosting service and
        invalid existing account due to mismatched Local Site
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example1.com',
            local_site=LocalSite.objects.create(name='test-site'))
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'myrepo',
            'bug_tracker_type': 'none',
        })
        form.validate_repository = False

        self.assertFalse(form.is_valid())
        self.assertFalse(form.hosting_account_linked)

    def test_with_hosting_service_custom_bug_tracker(self):
        """Testing RepositoryForm with a custom bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'custom',
            'bug_tracker': 'http://example.com/issue/%s',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertFalse(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker, 'http://example.com/issue/%s')
        self.assertNotIn('bug_tracker_type', repository.extra_data)

    def test_with_hosting_service_bug_tracker_service(self):
        """Testing RepositoryForm with a bug tracker service"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'test',
            'bug_tracker_hosting_account_username': 'testuser',
            'bug_tracker-test_repo_name': 'testrepo',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertFalse(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker,
                         'http://example.com/testuser/testrepo/issue/%s')
        self.assertEqual(repository.extra_data['bug_tracker_type'],
                         'test')
        self.assertEqual(
            repository.extra_data['bug_tracker-test_repo_name'],
            'testrepo')
        self.assertEqual(
            repository.extra_data['bug_tracker-hosting_account_username'],
            'testuser')

    def test_with_hosting_service_self_hosted_bug_tracker_service(self):
        """Testing RepositoryForm with a self-hosted bug tracker service"""
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'self_hosted_test',
            'bug_tracker_hosting_url': 'https://example.com',
            'bug_tracker-test_repo_name': 'testrepo',
        })
        form.validate_repository = False

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertFalse(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker,
                         'https://example.com/testrepo/issue/%s')
        self.assertEqual(repository.extra_data['bug_tracker_type'],
                         'self_hosted_test')
        self.assertEqual(
            repository.extra_data['bug_tracker-test_repo_name'],
            'testrepo')
        self.assertEqual(
            repository.extra_data['bug_tracker_hosting_url'],
            'https://example.com')

    def test_with_hosting_service_with_hosting_bug_tracker(self):
        """Testing RepositoryForm with hosting service's bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_use_hosting': True,
            'bug_tracker_type': 'googlecode',
        })
        form.validate_repository = False

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertTrue(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker,
                         'http://example.com/testuser/testrepo/issue/%s')
        self.assertNotIn('bug_tracker_type', repository.extra_data)
        self.assertFalse('bug_tracker-test_repo_name'
                         in repository.extra_data)
        self.assertFalse('bug_tracker-hosting_account_username'
                         in repository.extra_data)

    def test_with_hosting_service_with_hosting_bug_tracker_and_self_hosted(
            self):
        """Testing RepositoryForm with self-hosted hosting service's bug
        tracker
        """
        account = HostingServiceAccount.objects.create(
            username='testuser',
            service_name='self_hosted_test',
            hosting_url='https://example.com')
        account.data['password'] = 'testpass'
        account.save()

        account.data['authorization'] = {
            'token': '1234',
        }
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'self_hosted_test',
            'hosting_url': 'https://example.com',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_use_hosting': True,
            'bug_tracker_type': 'googlecode',
        })
        form.validate_repository = False

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertTrue(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker,
                         'https://example.com/testrepo/issue/%s')
        self.assertNotIn('bug_tracker_type', repository.extra_data)
        self.assertFalse('bug_tracker-test_repo_name'
                         in repository.extra_data)
        self.assertFalse('bug_tracker_hosting_url'
                         in repository.extra_data)

    def test_with_hosting_service_no_bug_tracker(self):
        """Testing RepositoryForm with no bug tracker"""
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        account.data['password'] = 'testpass'
        account.save()

        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
            'hosting_account': account.pk,
            'tool': self.git_tool_id,
            'test_repo_name': 'testrepo',
            'bug_tracker_type': 'none',
        })

        self.assertTrue(form.is_valid())

        repository = form.save()
        self.assertFalse(repository.extra_data['bug_tracker_use_hosting'])
        self.assertEqual(repository.bug_tracker, '')
        self.assertNotIn('bug_tracker_type', repository.extra_data)

    def test_with_hosting_service_with_existing_custom_bug_tracker(self):
        """Testing RepositoryForm with existing custom bug tracker"""
        repository = Repository(name='test',
                                bug_tracker='http://example.com/issue/%s')

        form = RepositoryForm(instance=repository)
        self.assertFalse(form._get_field_data('bug_tracker_use_hosting'))
        self.assertEqual(form._get_field_data('bug_tracker_type'), 'custom')
        self.assertEqual(form.initial['bug_tracker'],
                         'http://example.com/issue/%s')

    def test_with_hosting_service_with_existing_bug_tracker_service(self):
        """Testing RepositoryForm with existing bug tracker service"""
        repository = Repository(name='test')
        repository.extra_data['bug_tracker_type'] = 'test'
        repository.extra_data['bug_tracker-test_repo_name'] = 'testrepo'
        repository.extra_data['bug_tracker-hosting_account_username'] = \
            'testuser'

        form = RepositoryForm(instance=repository)
        self.assertFalse(form._get_field_data('bug_tracker_use_hosting'))
        self.assertEqual(form._get_field_data('bug_tracker_type'), 'test')
        self.assertEqual(
            form._get_field_data('bug_tracker_hosting_account_username'),
            'testuser')

        self.assertIn('test', form.bug_tracker_forms)
        self.assertIn('default', form.bug_tracker_forms['test'])
        bitbucket_form = form.bug_tracker_forms['test']['default']
        self.assertEqual(
            bitbucket_form.fields['test_repo_name'].initial,
            'testrepo')

    def test_with_hosting_service_with_existing_bug_tracker_using_hosting(
            self):
        """Testing RepositoryForm with existing bug tracker using hosting
        service
        """
        account = HostingServiceAccount.objects.create(username='testuser',
                                                       service_name='test')
        repository = Repository(name='test',
                                hosting_account=account)
        repository.extra_data['bug_tracker_use_hosting'] = True
        repository.extra_data['test_repo_name'] = 'testrepo'

        form = RepositoryForm(instance=repository)
        self.assertTrue(form._get_field_data('bug_tracker_use_hosting'))

    def test_bound_forms_with_post_with_repository_service(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted repository hosting_service using default plan
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'test',
        })

        # Make sure only the relevant forms are bound.
        for hosting_type, repo_forms in six.iteritems(form.repository_forms):
            for plan_id, repo_form in six.iteritems(repo_forms):
                self.assertEqual(repo_form.is_bound,
                                 hosting_type == 'test' and
                                 plan_id == form.DEFAULT_PLAN_ID)

        # Bug tracker info wasn't set in the form above.
        for hosting_type, bug_forms in six.iteritems(form.bug_tracker_forms):
            for plan_id, bug_form in six.iteritems(bug_forms):
                self.assertFalse(bug_form.is_bound)

        # Auth forms are never bound on initialize.
        for hosting_type, auth_form in six.iteritems(form.hosting_auth_forms):
            self.assertFalse(auth_form.is_bound)

    def test_bound_forms_with_post_with_bug_tracker_service(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted bug tracker hosting_service using default plan
        """
        form = RepositoryForm({
            'name': 'test',
            'bug_tracker_type': 'test',
        })

        # Make sure only the relevant forms are bound.
        for hosting_type, bug_forms in six.iteritems(form.bug_tracker_forms):
            for plan_id, bug_form in six.iteritems(bug_forms):
                self.assertEqual(bug_form.is_bound,
                                 hosting_type == 'test' and
                                 plan_id == form.DEFAULT_PLAN_ID)

        # Repository info wasn't set in the form above.
        for hosting_type, repo_forms in six.iteritems(form.repository_forms):
            for plan_id, repo_form in six.iteritems(repo_forms):
                self.assertFalse(repo_form.is_bound)

        # Auth forms are never bound on initialize.
        for hosting_type, auth_form in six.iteritems(form.hosting_auth_forms):
            self.assertFalse(auth_form.is_bound)

    def test_bound_forms_with_post_with_repo_service_and_plan(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted repository hosting_service with specific plans
        """
        form = RepositoryForm({
            'name': 'test',
            'hosting_type': 'github',
            'repository_plan': 'public',
        })

        # Make sure only the relevant forms are bound.
        for hosting_type, repo_forms in six.iteritems(form.repository_forms):
            for plan_id, repo_form in six.iteritems(repo_forms):
                self.assertEqual(repo_form.is_bound,
                                 hosting_type == 'github' and
                                 plan_id == 'public')

        # Bug tracker info wasn't set in the form above.
        for hosting_type, bug_forms in six.iteritems(form.bug_tracker_forms):
            for plan_id, bug_form in six.iteritems(bug_forms):
                self.assertFalse(bug_form.is_bound)

        # Auth forms are never bound on initialize.
        for hosting_type, auth_form in six.iteritems(form.hosting_auth_forms):
            self.assertFalse(auth_form.is_bound)

    def test_bound_forms_with_post_with_bug_tracker_service_and_plan(self):
        """Testing RepositoryForm binds hosting service forms only if matching
        posted bug tracker hosting_service with specific plans
        """
        form = RepositoryForm({
            'name': 'test',
            'bug_tracker_type': 'github',
            'bug_tracker_plan': 'public',
        })

        # Make sure only the relevant forms are bound.
        for hosting_type, bug_forms in six.iteritems(form.bug_tracker_forms):
            for plan_id, bug_form in six.iteritems(bug_forms):
                self.assertEqual(bug_form.is_bound,
                                 hosting_type == 'github' and
                                 plan_id == 'public')

        # Repository info wasn't set in the form above.
        for hosting_type, repo_forms in six.iteritems(form.repository_forms):
            for plan_id, repo_form in six.iteritems(repo_forms):
                self.assertFalse(repo_form.is_bound)

        # Auth forms are never bound on initialize.
        for hosting_type, auth_form in six.iteritems(form.hosting_auth_forms):
            self.assertFalse(auth_form.is_bound)



"""Repository support for Bazaar."""

from __future__ import unicode_literals

import os

import dateutil.parser
from django.utils import six
from django.utils.timezone import utc

from reviewboard.scmtools.core import SCMClient, SCMTool, HEAD, PRE_CREATION
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         InvalidRevisionFormatError,
                                         RepositoryNotFoundError, SCMError)
from reviewboard.ssh import utils as sshutils

try:
    import urlparse
    uses_netloc = urlparse.uses_netloc
except ImportError:
    import urllib.parse
    uses_netloc = urllib.parse.uses_netloc

# Register these URI schemes so we can handle them properly.
sshutils.ssh_uri_schemes.append('bzr+ssh')
uses_netloc.extend(['bzr', 'bzr+ssh'])


class BZRTool(SCMTool):
    """Repository support for Canonical's Bazaar.

    Bazaar is one of the first distributed version control systems, often
    used with the `Launchpad <https://launchpad.net>`_ service.

    Bazaar can be downloaded at http://bazaar-vcs.org/.
    """

    name = 'Bazaar'
    dependencies = {
        'executables': ['bzr'],
    }

    # Timestamp format in bzr diffs.
    # This isn't totally accurate: there should be a %z at the end.
    # Unfortunately, strptime() doesn't support %z.
    DIFF_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S'

    # "bzr diff" indicates that a file is new by setting the old
    # timestamp to the epoch time.
    PRE_CREATION_TIMESTAMP = '1970-01-01 00:00:00 +0000'

    REVISION_SPEC_KEYWORDS = (
        'ancestor:',
        'annotate:',
        'before:',
        'branch:',
        'date:',
        'last:',
        'mainline:',
        'revid:',
        'revno:',
        'submit:',
        'tag:',
    )

    def __init__(self, repository):
        """Initialize the Bazaar tool.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to communicate with.
        """
        super(BZRTool, self).__init__(repository)

        if repository.local_site:
            local_site_name = repository.local_site.name
        else:
            local_site_name = None

        self.client = BZRClient(path=repository.path,
                                local_site_name=local_site_name)

    def get_file(self, path, revision, **kwargs):
        """Return the contents from a file with the given path and revision.

        Args:
            path (unicode):
                The path of the file within the repository. This must not be
                a full Bazaar repository path.

            revision (unicode):
                The revision to fetch. If a Bazaar revision specifier keyword
                is provided, then it will be used to perform the lookup.
                Otherwise, this is assumed to be a date in the form of
                ``YYYY-MM-DD HH:MM:SS ZZZZ``, the format used in Bazaar diffs.

            **kwargs (dict, unused):
                Unused additional keyword arguments.

        Returns:
            bytes:
            The contents of the file from the repository.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file at the given revision was not found in the repostiory.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in a format that's not supported.
        """
        if revision == BZRTool.PRE_CREATION_TIMESTAMP:
            return b''

        revspec = self._revspec_from_revision(revision)

        if revspec is None:
            raise InvalidRevisionFormatError(path, revision)

        return self.client.get_file(path=path, revspec=revspec)

    def file_exists(self, path, revision, **kwargs):
        """Return whether a file exists with the given path and revision.

        Args:
            path (unicode):
                The path of the file within the repository. This must not be
                a full Bazaar repository path.

            revision (unicode):
                The revision to fetch. If a Bazaar revision specifier keyword
                is provided, then it will be used to perform the lookup.
                Otherwise, this is assumed to be a date in the form of
                ``YYYY-MM-DD HH:MM:SS ZZZZ``, the format used in Bazaar diffs.

            **kwargs (dict, unused):
                Unused additional keyword arguments.

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in a format that's not supported.
        """
        if revision == BZRTool.PRE_CREATION_TIMESTAMP:
            return False

        revspec = self._revspec_from_revision(revision)

        if revspec is None:
            raise InvalidRevisionFormatError(path, revision)

        return self.client.get_file_exists(path=path, revspec=revspec)

    def parse_diff_revision(self, file_str, revision_str, *args, **kwargs):
        """Parse a filename and revision identifier from a diff.

        If the revision identifer is a date indicating a new file, then
        this will return :py:data:`~reviewboard.scmtools.core.PRE_CREATION`.
        Otherwise, the revision identifier is returned directly.

        Args:
            file_str (unicode):
                The filename in the diff.

            revision_str (unicode):
                The revision in the diff.

            **kwargs (dict, unused):
                Unused additional keyword arguments.

        Returns:
            unicode:
            The revision identifier in the diff. If this is
            :py:attr:`~BZRTool.PRE_CREATION_TIMESTAMP`, then this will
            return :py:data:`~reviewboard.scmtools.core.PRE_CREATION`.
        """
        if revision_str == BZRTool.PRE_CREATION_TIMESTAMP:
            return (file_str, PRE_CREATION)

        return file_str, revision_str

    def _revspec_from_revision(self, revision):
        """Return a Bazaar revision specification based on the given revision.

        If the revision starts with a Bazaar revision specifier keyword
        argument, then the revision will be used as-is (allowing for the `bzr
        diff-revid <https://launchpad.net/bzr-diff-revid>`_ plugin to be used).

        Otherwise, this will attempt to match a date in
        ``YYYY-MM-DD HH:MM:SS ZZZZ` format (used by Bazaar diffs).

        Args:
            revision (unicode):
                The revision to parse.

        Returns:
            unicode:
            A revision specifier for the given revision. If a supported
            revision was not provided, this will return ``None``.
        """
        if revision == HEAD:
            revspec = 'last:1'
        elif revision.startswith(self.REVISION_SPEC_KEYWORDS):
            revspec = revision
        else:
            # Attempt to parse this as a timestamp into a Bazaar date revision
            # specifier.
            try:
                timestamp = dateutil.parser.parse(revision).astimezone(utc)
                revspec = 'date:%s' % timestamp.strftime('%Y-%m-%d,%H:%M:%S')
            except ValueError:
                revspec = None

        return revspec

    @classmethod
    def check_repository(cls, path, username=None, password=None,
                         local_site_name=None):
        """Check a repository to test its validity.

        This checks if a Bazaar repository exists and can be connected to. If
        the repository could not be found, an exception will be raised.

        Args:
            path (unicode):
                The repository path.

            username (unicode):
                The optional username used to connect to the repository.

            password (unicode):
                The optional password used to connect to the repository.

            local_site_name (unicode):
                The name of the Local Site that will own the repository.

        Raises:
            reviewboard.scmtools.errors.RepositoryNotFoundError:
                The repository could not be found, or there was an error
                communicating with it.
        """
        super(BZRTool, cls).check_repository(path, username, password,
                                             local_site_name)

        client = BZRClient(path=path,
                           local_site_name=local_site_name)

        if not client.is_valid_repository():
            raise RepositoryNotFoundError()


class BZRClient(SCMClient):
    """A client for performing Bazaar requests.

    This invokes the command line :command:`bzr` tool to perform file and
    repository lookups.
    """

    _bzr_plugin_path = None

    def __init__(self, path, local_site_name):
        """Initialize the client.

        Args:
            path (unicode):
                The repository path provided by the user.

            local_site_name (unicode):
                The name of the Local Site owning the repository.
        """
        if path.startswith('/'):
            self.path = 'file://%s' % path
        else:
            self.path = path

        self.local_site_name = local_site_name

    def is_valid_repository(self):
        """Return whether the provided repository information is valid.

        Returns:
            bool:
            ``True`` if information on the repository could be found.
            ``False`` if not.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                There was an error talking to Bazaar.
        """
        p = self._run_bzr(['info', self._build_repo_path(self.path)])
        errmsg = p.stderr.read()
        ret_code = p.wait()

        self._check_error(errmsg)

        return ret_code == 0

    def get_file(self, path, revspec):
        """Return the contents of a file.

        This expects a path within the repository and a Bazaar revision
        specifier.

        Args:
            path (unicode):
                The path to the file within the repository.

            revspec (unicode):
                The Bazaar revision specifier used to look up the file.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found.
        """
        path = self._build_repo_path(path)

        p = self._run_bzr(['cat', '-r', revspec, path])
        contents = p.stdout.read()
        errmsg = six.text_type(p.stderr.read())
        failure = p.wait()

        self._check_error(errmsg)

        if failure:
            raise FileNotFoundError(path=path,
                                    revision=revspec,
                                    detail=errmsg)

        return contents

    def get_file_exists(self, path, revspec):
        """Return whether a file exists in the repository.

        This expects a path within the repository and a Bazaar revision
        specifier.

        Args:
            path (unicode):
                The path to the file within the repository.

            revspec (unicode):
                The Bazaar revision specifier used to look up the file.

        Returns:
            bool:
            ``True`` if the file exists in the repository. ``False`` if not.
        """
        path = self._build_repo_path(path)
        p = self._run_bzr(['cat', '-r', revspec, path])
        errmsg = six.text_type(p.stderr.read())
        ret_code = p.wait()

        self._check_error(errmsg)

        return ret_code == 0

    def _run_bzr(self, args):
        """Run a Bazaar command.

        This will run :command:`bzr` with the specified arguments, and sets
        up the environment to work with :command:`rbssh`.

        Args:
            args (list of unicode):
                The list of arguments to pass to :command:`bzr`.

        Returns:
            subprocess.Popen:
            The handle for the process.
        """
        if not BZRClient._bzr_plugin_path:
            BZRClient._bzr_plugin_path = (
                '%s:%s' % (
                    os.path.join(os.path.dirname(__file__), 'plugins',
                                 'bzrlib', 'plugins'),
                    os.environ.get(b'BZR_PLUGIN_PATH', ''))
            ).encode('utf-8')

        return SCMTool.popen(
            ['bzr'] + args,
            local_site_name=self.local_site_name,
            env={
                b'BZR_PLUGIN_PATH': BZRClient._bzr_plugin_path,
                b'BZR_SSH': b'rbssh',
                b'TZ': b'UTC',
            })

    def _check_error(self, errmsg):
        """Check an error message from bzr and raise an exception if needed.

        If the error is an internal error, it will be raised, without the
        exception. If it's a known error that we can report better information
        on, then that information will be raised.

        Args:
            errmsg (unicode):
                The error message.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                A suitable error message, if an internal error was hit.
        """
        if 'Bazaar has encountered an internal error' in errmsg:
            if 'prefetch() takes exactly 2 arguments (1 given)' in errmsg:
                errmsg = ('Installed bzr and paramiko modules are '
                          'incompatible. See '
                          'https://bugs.launchpad.net/bzr/+bug/1524066')
            else:
                errmsg = errmsg.split(
                    'Traceback (most recent call last):')[0].strip()

            raise SCMError(errmsg)

    def _build_repo_path(self, path):
        """Return a path for a repository or file within a repository.

        The returned path is based on the repository path and the provided
        path within the repository. The resulting path can be passed to
        :py:meth:`_run_bzr`.

        Args:
            path (unicode):
                The path within the repository.

        Returns:
            unicode:
            The resulting repository path.
        """
        return '%s/%s' % (self.path, path.lstrip('/'))



"""Unit tests for reviewboard.accounts.decorators.valid_prefs_required."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpResponse, HttpResponseRedirect
from django.test.client import RequestFactory
from djblets.privacy.consent import (get_consent_requirements_registry,
                                     get_consent_tracker)

from reviewboard.accounts.decorators import valid_prefs_required
from reviewboard.accounts.models import Profile
from reviewboard.testing import TestCase


class ValidPrefsRequiredTests(TestCase):
    """Unit tests for reviewboard.accounts.decorators.valid_prefs_required."""

    @classmethod
    def setUpClass(cls):
        super(ValidPrefsRequiredTests, cls).setUpClass()

        cls.request_factory = RequestFactory()

    def setUp(self):
        super(ValidPrefsRequiredTests, self).setUp()

        self.user = User.objects.create(username='test-user')

        self.request = self.request_factory.get('/')
        self.request.user = self.user

    def test_with_anonymous_user(self):
        """Testing @valid_prefs_required with anonymous user"""
        self.request.user = AnonymousUser()

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = self._view_func(self.request)

        self.assertIs(type(response), HttpResponse)

    def test_with_consent_not_required(self):
        """Testing @valid_prefs_required with privacy_enable_user_consent=False
        """
        with self.siteconfig_settings({'privacy_enable_user_consent': False}):
            response = self._view_func(self.request)

        self.assertIs(type(response), HttpResponse)

    def test_with_consent_required_and_new_profile(self):
        """Testing @valid_prefs_required with privacy_enable_user_consent=True
        and new user profile
        """
        self.assertFalse(Profile.objects.filter(user=self.user).exists())

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = self._view_func(self.request)

        self.assertIs(type(response), HttpResponseRedirect)
        self.assertEqual(response.url, '/account/preferences/?next=/')

    def test_with_consent_required_and_consent_pending(self):
        """Testing @valid_prefs_required with privacy_enable_user_consent=True
        and pending consent
        """
        Profile.objects.create(user=self.user)

        consent_tracker = get_consent_tracker()
        all_consent = consent_tracker.get_all_consent(self.user)
        self.assertEqual(all_consent, {})

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = self._view_func(self.request)

        self.assertIs(type(response), HttpResponseRedirect)
        self.assertEqual(response.url, '/account/preferences/?next=/')

    def test_with_consent_required_and_no_consent_pending(self):
        """Testing @valid_prefs_required with privacy_enable_user_consent=True
        and no pending consent
        """
        Profile.objects.create(user=self.user)

        consent_tracker = get_consent_tracker()
        consent_tracker.record_consent_data_list(
            self.user,
            [
                consent_requirement.build_consent_data(granted=True)
                for consent_requirement in get_consent_requirements_registry()
            ])

        all_consent = consent_tracker.get_all_consent(self.user)
        self.assertNotEqual(all_consent, {})

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = self._view_func(self.request)

        self.assertIs(type(response), HttpResponse)

    def test_with_consent_required_pending_consent_enabled_decorator(self):
        """Testing @valid_prefs_required with disbled_consent_checks= set to a
        function that always returns False
        """
        @valid_prefs_required(disable_consent_checks=lambda request: False)
        def view_func(request):
            return HttpResponse()

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = view_func(self.request)

        self.assertIs(type(response), HttpResponseRedirect)

    def test_with_consent_required_pending_consent_disabled_decorator(self):
        """Testing @valid_prefs_required with disbled_consent_checks= set to a
        function that always returns True
        """
        @valid_prefs_required(disable_consent_checks=lambda request: True)
        def view_func(request):
            return HttpResponse()

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = view_func(self.request)

        self.assertIs(type(response), HttpResponse)

    def test_with_consent_required_pending_consent_decorator_function(self):
        """Testing @valid_prefs_required with disbled_consent_checks= set to a
        function
        """
        def disable_consent_checks(request):
            return 'disable-consent-checks' in request.GET

        @valid_prefs_required(disable_consent_checks=disable_consent_checks)
        def view_func(request):
            return HttpResponse()

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = view_func(self.request)

        self.assertIs(type(response), HttpResponseRedirect)

        request = self.request_factory.get('/?disable-consent-checks')
        request.user = self.user

        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            response = view_func(request)

        self.assertIs(type(response), HttpResponse)

    @staticmethod
    @valid_prefs_required
    def _view_func(request):
        return HttpResponse()



"""Tests for the ReviewRequestLastUpdateResource."""

from __future__ import unicode_literals

from dateutil.parser import parse as parse_date
from django.contrib.auth.models import User
from django.utils import six
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.models import (Review, ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import \
    review_request_last_update_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import get_review_request_last_update_url


@six.add_metaclass(BasicTestsMetaclass)
class ResourceTests(BaseWebAPITestCase):
    """Testing ReviewRequestLastUpdateResource APIs."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/last-update/'
    test_http_methods = ('GET',)
    resource = resources.review_request_last_update

    def compare_item(self, item_rsp, review_request, expected_user=None):
        """Compare a response to a review request.

        Args:
            item_rsp (dict):
                A dictionary of serialized data from the resource.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request whose latest activity will be used for
                comparison.

            expected_user (django.contrib.auth.models.User, optional):
                The user we expected to be serialized in ``item_rsp``.

                If not provided, this defaults to ``review_request.submitter``.

        Raises:
            AssertionError:
                The serialized response does not correspond with the given
                review request or user.
        """
        if expected_user is None:
            expected_user = review_request.submitter

        item_type = item_rsp['type']
        last_updated = review_request.last_updated.replace(microsecond=0)

        last_activity = review_request.get_last_activity_info()
        updated_object = last_activity['updated_object']
        changedesc = last_activity['changedesc']

        rsp_timestamp = parse_date(item_rsp['timestamp']).replace(
            microsecond=0)

        if item_type == 'review-request':
            self.assertIsInstance(updated_object, ReviewRequest)
            self.assertEqual(updated_object, review_request)
            object_timestamp = updated_object.last_updated
        elif item_type == 'diff':
            self.assertIsInstance(updated_object, DiffSet)
            object_timestamp = updated_object.timestamp
        elif item_type in ('review', 'reply'):
            self.assertIsInstance(updated_object, Review)
            object_timestamp = updated_object.timestamp

            self.assertIsNone(last_activity['changedesc'])
        else:
            self.fail('Unknown item type "%s"' % item_type)

        object_timestamp = object_timestamp.replace(microsecond=0)

        self.assertEqual(last_updated, object_timestamp)
        self.assertEqual(last_updated, rsp_timestamp)

        if changedesc:
            self.assertEqual(expected_user,
                             changedesc.get_user(review_request))

        self.assertEqual(expected_user.pk,
                         item_rsp['user']['id'])

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site=False,
                             local_site_name=None, update_review_request=False,
                             update_diffset=False, update_review=False,
                             update_reply=False, publish_user=None):
        """Setup a basic HTTP GET test.

        Args:
            user (django.contrib.auth.models.User):
                The user that will be used to retrieve the resource.

                They will also be the submitter/author of all objects (although
                not necessary the publisher).

            with_local_site (boolean, optional):
                Whether or not a LocalSite-specific resource will be used.

            local_site_name (unicode, optional):
                The name of the LocalSite to use.

            update_review_request (boolean, optional):
                Whether or not the review request should be updated after
                publishing.

            update_diffset (boolean, optional):
                Whether or not the review request should be updated with a diff
                after publishing.

            update_review (boolean, optional):
                Whether or not the review request should be reviewed after
                publishing.

            update_reply (boolean, optional):
                Whther or not to create a reply.

                This implies ``update_review``.

            publish_user (django.contrib.auth.models.User, optional):
                The user that will trigger the update (i.e., they will publish
                all requested objects).

                If not provided, ``user`` will be used.

        Returns:
            tuple:
            A 3-tuple of:

            * The URL to request.
            * The expected mimetype.
            * The review request to use for comparison.
        """
        update_review = update_review or update_reply
        publish_user = publish_user or user
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user,
            with_local_site=with_local_site,
            target_people=[user])

        review_request.publish(user=publish_user)

        if update_review_request:
            draft = ReviewRequestDraft.create(review_request)
            draft.summary = '%s updated' % review_request.summary
            draft.save()

            review_request.publish(user=publish_user)

        if update_diffset:
            self.create_diffset(review_request=review_request, draft=True)
            review_request.publish(user=publish_user)

        if update_review:
            review = self.create_review(review_request=review_request,
                                        publish=False)
            review.publish(user=publish_user)

            if update_reply:
                reply = self.create_reply(review)
                reply.publish(user=publish_user)

        return (
            get_review_request_last_update_url(local_site_name=local_site_name,
                                               review_request=review_request),
            review_request_last_update_mimetype,
            review_request,
        )

    @webapi_test_template
    def test_get_other_user(self):
        """Testing the GET <URL> API when the review request has been published
        by another user
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review-request')

        # We have no way of knowing what user published based on only the
        # model, so it will always be the submitter.
        self.compare_item(item_rsp, review_request, expected_user=self.user)

    @webapi_test_template
    def test_get_review_request_updated(self):
        """Testing the GET <URL> API when the review request has been updated
        """
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_review_request=True)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review-request')
        self.compare_item(item_rsp, review_request)

    @webapi_test_template
    def test_get_review_request_updated_other_user(self):
        """Testing the GET <URL> API when a review request has been updated
        by a different user
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_review_request=True,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review-request')
        self.compare_item(item_rsp, review_request, expected_user=publish_user)

    @webapi_test_template
    def test_get_diff_updated(self):
        """Testing the GET <URL> API when a new diff has been published"""
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_diffset=True)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'diff')
        self.compare_item(rsp['last_update'], review_request)

    @webapi_test_template
    def test_get_diff_updated_other_user(self):
        """Testing the GET <URL> API when a new diff has been published by
        another user
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_diffset=True,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'diff')
        self.compare_item(item_rsp, review_request, expected_user=publish_user)

    @webapi_test_template
    def test_get_review_published(self):
        """Testing the GET <URL> API when a review has been published"""
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_review=True)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review')
        self.compare_item(item_rsp, review_request)

    @webapi_test_template
    def test_get_review_published_other_user(self):
        """Testing the GET <URL> API when a review has been published by
        another user
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_review=True,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review')

        # We have no way of knowing what user published based on only the
        # model, so it will always be the review author.
        self.compare_item(item_rsp, review_request, expected_user=self.user)

    @webapi_test_template
    def test_get_reply_published(self):
        """Testing the GET <URL> API when a review reply has been published"""
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_reply=True)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'reply')
        self.compare_item(item_rsp, review_request)

    @webapi_test_template
    def test_get_reply_published_other_user(self):
        """Testing the GET <URL> API when a review reply has been published by
        another user
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_reply=True,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'reply')

        # We have no way of knowing what user published based on only the
        # model, so it will always be the review author.
        self.compare_item(item_rsp, review_request, expected_user=self.user)

    @webapi_test_template
    def test_get_review_request_updated_review_published_other_user(self):
        """Testing the GET <URL> API when a review request has been updated
        and a review request has been published
        """
        publish_user = User.objects.get(username='admin')
        url, expected_mimetype, review_request = self.setup_basic_get_test(
            self.user,
            update_review_request=True,
            update_review=True,
            publish_user=publish_user)
        rsp = self.api_get(url, expected_mimetype=expected_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['last_update']

        self.assertEqual(item_rsp['type'], 'review')

        # We have no way of knowing what user published based on only the
        # model, so it will always be the review author.
        self.compare_item(item_rsp, review_request, expected_user=self.user)



from __future__ import unicode_literals

import re

from django.utils import six
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)

from reviewboard.notifications.forms import WebHookTargetForm
from reviewboard.notifications.models import WebHookTarget
from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.decorators import (webapi_login_required,
                                           webapi_check_local_site,
                                           webapi_response_errors)
from reviewboard.webapi.mixins import UpdateFormMixin
from reviewboard.webapi.resources.repository import RepositoryResource


class WebHookResource(UpdateFormMixin, WebAPIResource):
    """A resource for managing webhooks.

    Webhooks are HTTP-based callbacks. When events are triggered on Review
    Board (such as a review request being published), a webhook can be
    configured to make a request to a URL with a payload.

    Webhooks can be triggered by the following events:

    * Review requests being published (``review_request_published``).
    * Review requests being closed (``review_request_closed``).
    * Reviews being published (``review_published``).
    * Review replies being published (``reply_published``).
    * All of the above (``*``).

    The payload that is sent depends on the type of event. Review Board will
    generate a default payload for each event, or it can be overridden by
    providing your payload as the ``custom_content`` field.

    By default, the generated payload will be a JSON payload. This can be
    changed by setting the ``encoding`` field to the appropriate value. Valid
    encodings are:

    * ``application/json``
    * ``application/xml``
    * ``application/x-www-form-encoded``

    Payloads can also be signed with a 128-byte secret given in the ``secret``
    field. If provided, the payload will be signed with the HMAC algorithm.

    Webhooks can apply to all repositories, a select number of repositories, or
    to no repositories. In the latter case, they will only trigger with
    attachment-only review requests.

    Webhooks are :term:`Local Site`-dependant. They will only trigger for
    repositories in the configured local site. If no Local Site is configured,
    they will apply to review requests without a Local Site. If a webhook is
    configured to trigger for a specific set of repositories, the webhook and
    all repositories must be in the same Local Site.
    """

    added_in = '2.5'

    name = 'webhook'
    verbose_name = 'WebHook'
    model = WebHookTarget
    uri_object_key = 'webhook_id'
    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')

    form_class = WebHookTargetForm

    APPLY_TO_ALL_REPOS = 'all'
    APPLY_TO_NO_REPOS = 'none'
    APPLY_TO_CUSTOM_REPOS = 'custom'

    ALL_APPLY_TO_OPTIONS = (
        APPLY_TO_ALL_REPOS,
        APPLY_TO_NO_REPOS,
        APPLY_TO_CUSTOM_REPOS,
    )

    COMMA_SPLIT_RE = re.compile(r'\s*,\s*')

    fields = {
        'apply_to': {
            'type': ALL_APPLY_TO_OPTIONS,
            'description': 'What review requests the webhook applies to. This '
                           'is one of the strings ``all``, ``none``, or '
                           '``custom``. In the case of ``custom``, the '
                           'repositories are specified in the '
                           '``repositories`` field.',
        },
        'custom_content': {
            'type': six.text_type,
            'description': 'An optional custom payload.',
        },
        'enabled': {
            'type': bool,
            'description': 'Whether or not the webhook is enabled.',
        },
        'encoding': {
            'type': WebHookTarget.ALL_ENCODINGS,
            'description': 'The encoding for the payload. This is one of '
                           '``application/json``, ``application/xml`` or '
                           '``application/x-www-form-data``.',
        },
        'events': {
            'type': [six.text_type],
            'description': 'A list of events that will cause the webhook to '
                           'trigger.',
        },
        'extra_data': {
            'type': dict,
            'description': 'Extra data as part of the webhook. '
                           'This can be set by the API or extensions.',
        },
        'id': {
            'type': int,
            'description': 'The numeric ID of the webhook.',
        },
        'secret': {
            'type': six.text_type,
            'description': 'An optional HMAC digest for the webhook payload. '
                           'If this is specified, the payload will be signed '
                           'with it.',
        },
        'repositories': {
            'type': [RepositoryResource],
            'description': 'The list of repositories this applies to.',
        },
        'url': {
            'type': six.text_type,
            'description': 'The URL to make HTTP requests against.',
        },
    }

    def has_list_access_permissions(self, request, local_site=None, *args,
                                    **kwargs):
        """Determine if the user has list access permissions for the resource.

        Only superusers and local site admins have access to this resource.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The current local site if it exists.

            *args (list):
                Extra arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            bool: Whether or not the user has list access permissions.
        """
        return WebHookTarget.objects.can_create(request.user, local_site)

    def has_access_permissions(self, request, obj, local_site=None, *args,
                               **kwargs):
        """Determine if the user has access permissions for the resource.

        Only superusers and local site admins have access to the resource.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The current local site if it exists.

            *args (list):
                Extra arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            bool: Whether or not the user has access permissions.
        """
        return obj.is_accessible_by(request.user, local_site=local_site)

    def has_modify_permissions(self, *args, **kwargs):
        """Determine if the user has modify permissions for the resource.

        Only admins have access to this resource.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The current local site if it exists.

            *args (list):
                Extra arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            bool: Whether or not the user has modify permissions.
        """
        return self.has_access_permissions(*args, **kwargs)

    def has_delete_permissions(self, *args, **kwargs):
        """Determine if the user has modification permissions for the resource.

        Only superusers and local site admins have modification permissions.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The current local site if it exists.

            *args (list):
                Extra arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            bool: Whether or not the user has delete permissions.
        """
        return self.has_access_permissions(*args, **kwargs)

    def get_queryset(self, request, local_site=None, *args, **kwargs):
        """Return the queryset for all objects that match the given request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            local_site (reviewboard.site.models.LocalSite):
                The current local site if it exists.

            *args (list):
                Extra arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            The queryset for all objects matching the given request.
        """
        return WebHookTarget.objects.for_local_site(local_site)

    def serialize_apply_to_field(self, obj, *args, **kwargs):
        """Serialize the ``apply_to`` field into a human-readable value.

        Args:
            value (unicode):
                The value to parse.

            request (django.http.HttpRequest):
                The HTTP request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            unicode: A human readable value for the field.
        """
        if obj.apply_to == WebHookTarget.APPLY_TO_ALL:
            return 'all'
        elif obj.apply_to == WebHookTarget.APPLY_TO_NO_REPOS:
            return 'none'
        elif obj.apply_to == WebHookTarget.APPLY_TO_SELECTED_REPOS:
            return 'custom'
        else:
            assert False

    def parse_apply_to_field(self, value, request, **kwargs):
        """Parse the apply_to field from a human-readable value.

        Args:
            value (unicode):
                The value to parse.

            request (django.http.HttpRequest):
                The HTTP request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            unicode:
            A value that the ``WebHookTarget`` model expects for its
            ``apply_to`` field.
        """
        value = value.lower()

        if value == 'all':
            return WebHookTarget.APPLY_TO_ALL
        elif value == 'custom':
            return WebHookTarget.APPLY_TO_SELECTED_REPOS
        elif value == 'none':
            return WebHookTarget.APPLY_TO_NO_REPOS
        else:
            assert False

    def parse_repositories_field(self, value, request, **kwargs):
        """Parse the comma-separated repository IDs.

        Args:
            value (unicode):
                Comma-separated integers representing repository IDs.

            request (django.http.HttpRequest):
                The HTTP request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            list: A list of repository IDs as strings.
        """
        return self._parse_comma_list(value)

    def parse_events_field(self, value, request, **kwargs):
        """Parse the comma-separated event names.

        Args:
            value (unicode):
                Comma-separated event names.

            request (django.http.HttpRequest):
                The HTTP request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            list: A list of event names as strings.
        """
        return self._parse_comma_list(value)

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Retrieves information about a webhook.

        This endpoint will retrieve all information pertaining to the requested
        webhook.
        """
        pass

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def get_list(self, *args, **kwargs):
        """Retrieves the list of webhooks."""
        pass

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def delete(self, *args, **kwargs):
        """Deletes a webhook."""
        pass

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED)
    @webapi_request_fields(
        required={
            'apply_to': {
                'type': ALL_APPLY_TO_OPTIONS,
                'description': 'What review requests the webhook applies to. '
                               'This is one of the strings ``all``, ``none``, '
                               'or ``custom``. In the case of ``custom``, the '
                               'repositories should be specified in the '
                               '``repositories`` field.',
            },
            'enabled': {
                'type': bool,
                'description': 'Whether or not the webhook is enabled.',
            },
            'encoding': {
                'type': WebHookTarget.ALL_ENCODINGS,
                'description': 'The encoding for the payload. This is one of '
                               '``application/json``, ``application/xml`` or '
                               '``application/x-www-form-encoded``.',
            },
            'events': {
                'type': six.text_type,
                'description': 'The type of events that trigger the webhook. '
                               'This should be a list of values separated by '
                               'commas.',
            },
            'url': {
                'type': six.text_type,
                'description': 'The URL to make HTTP requests against.',
            },
        },
        optional={
            'custom_content': {
                'type': six.text_type,
                'description': 'An optional custom payload.',
            },
            'secret': {
                'type': six.text_type,
                'description': 'An optional HMAC digest for the webhook '
                               'payload. If this is specified, the payload '
                               'will be signed with it.',
            },
            'repositories': {
                'type': six.text_type,
                'description': 'If ``apply_to`` is ``selected repositories``, '
                               'this is a comma-separated list of repository '
                               'IDs that the webhook applies to.',
            },
        },
        allow_unknown=True
    )
    def create(self, request, parsed_request_fields, local_site=None,
               extra_fields=None, *args, **kwargs):
        """Creates a new webhook.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.

        Extra data values supplied will not be used when building the payload
        of the webhook.
        """
        if not WebHookTarget.objects.can_create(request.user, local_site):
            return self.get_no_access_error(request)

        form_data = parsed_request_fields.copy()

        if local_site:
            form_data['local_site'] = local_site.pk

        custom_content = form_data.get('custom_content', '')
        form_data['use_custom_content'] = (custom_content != '')

        form = self.create_form(form_data, request)

        if form.is_valid():
            try:
                instance = self.save_form(form, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

            return 201, {
                self.item_result_key: instance,
            }
        else:
            return INVALID_FORM_DATA, {
                'fields': self._get_form_errors(form),
            }

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED)
    @webapi_request_fields(
        optional={
            'apply_to': {
                'type': ALL_APPLY_TO_OPTIONS,
                'description': 'What review requests the webhook applies to. '
                               'This is one of the strings ``all``, ``none``, '
                               'or ``custom``. In the case of ``custom``, the '
                               'repositories should be specified in the '
                               '``repositories`` field.',
            },
            'custom_content': {
                'type': six.text_type,
                'description': 'An optional custom payload.',
            },
            'enabled': {
                'type': bool,
                'description': 'Whether or not the webhook is enabled.',
            },
            'encoding': {
                'type': WebHookTarget.ALL_ENCODINGS,
                'description': 'The encoding for the payload. This is one of '
                               '``application/json``, ``application/xml`` or '
                               '``application/x-www-form-encoded``.',
            },
            'events': {
                'type': six.text_type,
                'description': 'The type of events that trigger the webhook. '
                               'This should be a list of values separated by '
                               'commas.',
            },
            'url': {
                'type': six.text_type,
                'description': 'The URL to make HTTP requests against.',
            },
            'secret': {
                'type': six.text_type,
                'description': 'An optional HMAC digest for the webhook '
                               'payload. If this is specified, the payload '
                               'will be signed with it.',
            },
            'repositories': {
                'type': six.text_type,
                'description': 'If ``apply_to`` is ``selected repositories``, '
                               'this is a comma-separated list of repository '
                               'IDs that the webhook applies to.',
            },
        },
        allow_unknown=True
    )
    def update(self, request, parsed_request_fields, local_site=None,
               extra_fields=None, *args, **kwargs):
        """Updates a webhook.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.

        Extra data values supplied will not be used when building the payload
        of the webhook.
        """
        try:
            webhook = self.get_object(request, local_site=local_site, *args,
                                      **kwargs)
        except WebHookTarget.DoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_modify_permissions(request, webhook,
                                           local_site=local_site):
            return self.get_no_access_error(request)

        form_data = parsed_request_fields.copy()

        if local_site:
            form_data['local_site'] = local_site.pk

        if 'custom_content' in form_data:
            # We only explicitly set use_custom_content if the user has
            # provided the custom_content field. We don't want to unset it
            # when the user is updating other fields and does not intend to
            # update this one.
            form_data['use_custom_content'] = \
                (form_data['custom_content'] != '')

        form = self.create_form(form_data, request, instance=webhook)

        if form.is_valid():
            try:
                instance = self.save_form(form, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

            return 200, {
                self.item_result_key: instance,
            }
        else:
            return INVALID_FORM_DATA, {
                'fields': self._get_form_errors(form),
            }

    def _parse_comma_list(self, value):
        """Split a comma-separated string.

        Args:
            value (unicode):
                The comma-separated list of values.

        Returns:
            list:
            A list of :py:class:`unicode` objects. If the given value is empty,
            then the empty list is returned instead of a list containing the
            empty string.
        """
        value = value.strip()

        if not value:
            return []

        return self.COMMA_SPLIT_RE.split(value)


webhook_resource = WebHookResource()



from __future__ import unicode_literals

from django.contrib.auth.models import User
from djblets.auth.signals import user_registered
from kgb import SpyAgency

from reviewboard.accounts.models import _add_default_groups
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class DefaultGroupTest(SpyAgency, TestCase):
    """Unit tests for default group memberships."""

    fixtures = ['test_users', 'test_site']

    def test_user_registration(self):
        """Testing default group computation on user_registered signal"""
        self.spy_on(_add_default_groups)

        user = User.objects.create_user(username='reviewboard',
                                        email='user@example.com',
                                        password='password')

        user_registered.send(sender=None, user=user)

        self.assertTrue(_add_default_groups.called)
        self.assertTrue(_add_default_groups.last_called_with(user=user,
                                                             local_site=None))

    def test_local_site_add_user(self):
        """Testing default group computation when adding user to LocalSite
        users list
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.get(pk=3)

        self.spy_on(_add_default_groups)

        local_site.users.add(user)

        self.assertTrue(_add_default_groups.called)
        self.assertTrue(_add_default_groups.last_called_with(
            user=user,
            local_site=local_site))

    def test_user_add_local_site(self):
        """Testing default group computation when adding LocalSite to user's
        LocalSite list
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.get(pk=3)

        self.spy_on(_add_default_groups)

        user.local_site.add(local_site)

        self.assertTrue(_add_default_groups.spy.called)
        self.assertTrue(_add_default_groups.spy.last_called_with(
            user=user,
            local_site=local_site))

    def test_add_default_groups(self):
        """Testing default group computation without LocalSite"""
        user = User.objects.get(pk=1)
        group_count_before = user.review_groups.count()

        _add_default_groups(sender=None, user=user)

        self.assertEqual(group_count_before,
                         User.objects.get(pk=user.id).review_groups.count())

        self.create_review_group(is_default_group=True)

        _add_default_groups(sender=None, user=user)

        self.assertEqual(group_count_before + 1,
                         User.objects.get(pk=user.id).review_groups.count())

    def test_add_default_groups_with_local_site(self):
        """Testing default group computation with LocalSite"""
        user = User.objects.get(pk=3)
        local_site = LocalSite.objects.create(name='test')
        self.create_review_group(is_default_group=True, local_site=local_site)

        group_count_before = user.review_groups.count()

        _add_default_groups(sender=None, user=user, local_site=local_site)

        self.assertEqual(group_count_before + 1,
                         User.objects.get(pk=user.id).review_groups.count())



"""Unit tests for reviewboard.accounts.mixins.UserProfileRequiredViewMixin."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpResponse
from django.test.client import RequestFactory
from django.views.generic.base import View

from reviewboard.accounts.mixins import UserProfileRequiredViewMixin
from reviewboard.accounts.models import Profile
from reviewboard.testing import TestCase


class UserProfileRequiredViewMixinTests(TestCase):
    """Unit tests for reviewboard.accounts.mixins.UserProfileRequiredViewMixin.
    """

    def test_dispatch_with_no_profile(self):
        """Testing UserProfileRequiredViewMixin.dispatch with authenticated
        user without a profile
        """
        class MyView(UserProfileRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNotNone(view.request.user.get_profile())

                return HttpResponse('success')

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='doc',
                                                email='doc@example.com')

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')

    def test_dispatch_with_profile(self):
        """Testing UserProfileRequiredViewMixin.dispatch with authenticated
        user with a profile
        """
        class MyView(UserProfileRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNotNone(view.request.user.get_profile())

                return HttpResponse('success')

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='doc',
                                                email='doc@example.com')
        Profile.objects.create(user=request.user)

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')

    def test_dispatch_with_anonymous(self):
        """Testing UserProfileRequiredViewMixin.dispatch with anonymous user"""
        class MyView(UserProfileRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsInstance(view.request.user, AnonymousUser)

                return HttpResponse('success')

        request = RequestFactory().request()
        request.user = AnonymousUser()

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')



from __future__ import unicode_literals

from django.utils import six
from django.utils.six.moves.urllib.parse import (parse_qs, urlencode,
                                                 urlsplit, urlunsplit)


class InvalidPageError(Exception):
    """An error representing an invalid page access."""
    pass


class BasePaginator(object):
    """Base class for a paginator used in the hosting services code.

    This provides the basic state and stubbed functions for a simple
    paginator. Subclasses can build upon this to offer more advanced
    functionality.
    """
    def __init__(self, start=None, per_page=None):
        self.start = start
        self.per_page = per_page
        self.page_data = None
        self.total_count = None

    @property
    def has_prev(self):
        """Returns whether there's a previous page available.

        Subclasses must override this to provide a meaningful
        return value.
        """
        raise NotImplementedError

    @property
    def has_next(self):
        """Returns whether there's a next page available.

        Subclasses must override this to provide a meaningful
        return value.
        """
        raise NotImplementedError

    def prev(self):
        """Fetches the previous page, returning the page data.

        Subclasses must override this to provide the logic for
        fetching pages.

        If there isn't a previous page available, this must raise
        InvalidPageError.
        """
        raise NotImplementedError

    def next(self):
        """Fetches the previous page, returning the page data.

        Subclasses must override this to provide the logic for
        fetching pages.

        If there isn't a next page available, this must raise
        InvalidPageError.
        """
        raise NotImplementedError


class APIPaginator(BasePaginator):
    """Handles pagination for API requests to a hosting service.

    Hosting services may provide subclasses of APIPaginator that can
    handle paginating their specific APIs. These make it easy to fetch
    pages of data from the API, and also works as a bridge for
    Review Board's web API resources.

    All APIPaginators are expected to take an instance of a
    HostingServiceClient subclass, and the starting URL (without any
    arguments for pagination).

    Subclasses can access the HostingServiceClient through the ``client``
    member of the paginator in order to perform requests against the
    HostingService.
    """
    #: The optional query parameter name used to specify the start page in
    #: a request.
    start_query_param = None

    #: The optional query parameter name used to specify the requested number
    #: of results per page.
    per_page_query_param = None

    def __init__(self, client, url, query_params={}, *args, **kwargs):
        super(APIPaginator, self).__init__(*args, **kwargs)

        self.client = client
        self.prev_url = None
        self.next_url = None
        self.page_headers = None

        # Augment the URL with the provided query parameters.
        query_params = query_params.copy()

        if self.start_query_param and self.start:
            query_params[self.start_query_param] = self.start

        if self.per_page_query_param and self.per_page:
            query_params[self.per_page_query_param] = self.per_page

        self.url = self._add_query_params(url, query_params)

        self._fetch_page()

    @property
    def has_prev(self):
        """Returns whether there's a previous page available."""
        return self.prev_url is not None

    @property
    def has_next(self):
        """Returns whether there's a next page available."""
        return self.next_url is not None

    def prev(self):
        """Fetches the previous page, returning the page data.

        If there isn't a next page available, this will raise
        InvalidPageError.
        """
        if not self.has_prev:
            raise InvalidPageError

        self.url = self.prev_url
        return self._fetch_page()

    def next(self):
        """Fetches the next page, returning the page data.

        If there isn't a next page available, this will raise
        InvalidPageError.
        """
        if not self.has_next:
            raise InvalidPageError

        self.url = self.next_url
        return self._fetch_page()

    def fetch_url(self, url):
        """Fetches the URL, returning information on the page.

        This must be implemented by subclasses. It must return a dictionary
        with the following fields:

        * data        - The data from the page (generally as a list).
        * headers     - The headers from the page response.
        * total_count - The optional total number of items across all pages.
        * per_page    - The optional limit on the number of items fetched
                        on each page.
        * prev_url    - The optional URL to the previous page.
        * next_url    - The optional URL to the next page.
        """
        raise NotImplementedError

    def _fetch_page(self):
        """Fetches a page and extracts the information from it."""
        page_info = self.fetch_url(self.url)

        self.prev_url = page_info.get('prev_url')
        self.next_url = page_info.get('next_url')
        self.per_page = page_info.get('per_page', self.per_page)
        self.page_data = page_info.get('data')
        self.page_headers = page_info.get('headers')
        self.total_count = page_info.get('total_count')

        return self.page_data

    def _add_query_params(self, url, new_query_params):
        """Adds query parameters onto the given URL."""
        scheme, netloc, path, query_string, fragment = urlsplit(url)
        query_params = parse_qs(query_string)
        query_params.update(new_query_params)
        new_query_string = urlencode(
            [
                (key, value)
                for key, value in sorted(six.iteritems(query_params),
                                         key=lambda i: i[0])
            ],
            doseq=True)

        return urlunsplit((scheme, netloc, path, new_query_string, fragment))


class ProxyPaginator(BasePaginator):
    """A paginator that proxies to another paginator, transforming data.

    This attaches to another paginator, forwarding all requests and proxying
    all data.

    The ProxyPaginator can take the data returned from the other paginator
    and normalize it, transforming it into a new form.

    This is useful when a HostingService wants to return a paginator to
    callers that represents data in a structured way, using an APIPaginator's
    raw payloads as a backing.
    """
    def __init__(self, paginator, normalize_page_data_func=None):
        # NOTE: We're not calling BasePaginator here, because we're actually
        #       overriding all the properties it would set that we care about.
        self.paginator = paginator
        self.normalize_page_data_func = normalize_page_data_func
        self.page_data = self.normalize_page_data(self.paginator.page_data)

    @property
    def has_prev(self):
        """Returns whether there's a previous page available."""
        return self.paginator.has_prev

    @property
    def has_next(self):
        """Returns whether there's a next page available."""
        return self.paginator.has_next

    @property
    def per_page(self):
        """Returns the number of items requested per page."""
        return self.paginator.per_page

    @property
    def total_count(self):
        """Returns the number of items across all pages, if known."""
        return self.paginator.total_count

    def prev(self):
        """Fetches the previous page, returning the page data.

        If there isn't a next page available, this will raise
        InvalidPageError.
        """
        return self._process_page(self.paginator.prev())

    def next(self):
        """Fetches the next page, returning the page data.

        If there isn't a next page available, this will raise
        InvalidPageError.
        """
        return self._process_page(self.paginator.next())

    def normalize_page_data(self, data):
        """Normalizes a page of data.

        If ``normalize_page_data_func`` was passed on construction, this
        will call it, passing in the page data. That will then be returned.

        This can be overridden by subclasses that want to do more complex
        processing without requiring ``normalize_page_data_func`` to be
        passed in.
        """
        if callable(self.normalize_page_data_func):
            data = self.normalize_page_data_func(data)

        return data

    def _process_page(self, page_data):
        """Processes a page of data.

        This will normalize the page data, store it, and return it.
        """
        self.page_data = self.normalize_page_data(page_data)

        return self.page_data



"""Review Board e-mail module."""

from __future__ import absolute_import, unicode_literals

import email

from django.db.models.signals import post_delete
from djblets.auth.signals import user_registered

from reviewboard.notifications.email.signal_handlers import (
    send_reply_published_mail,
    send_review_published_mail,
    send_review_request_closed_mail,
    send_review_request_published_mail,
    send_user_registered_mail,
    send_webapi_token_created_mail,
    send_webapi_token_deleted_mail,
    send_webapi_token_updated_mail)
from reviewboard.notifications.email.hooks import (register_email_hook,
                                                   unregister_email_hook)
from reviewboard.reviews.models import ReviewRequest, Review
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)
from reviewboard.webapi.models import WebAPIToken
from djblets.webapi.signals import webapi_token_created, webapi_token_updated

def connect_signals():
    """Connect e-mail callbacks to signals."""
    signal_table = [
        (reply_published, send_reply_published_mail, Review),
        (review_published, send_review_published_mail, Review),
        (review_request_closed, send_review_request_closed_mail,
         ReviewRequest),
        (review_request_published, send_review_request_published_mail,
         ReviewRequest),
        (user_registered, send_user_registered_mail, None),
        (webapi_token_created, send_webapi_token_created_mail, WebAPIToken),
        (webapi_token_updated, send_webapi_token_updated_mail, WebAPIToken),
        (post_delete, send_webapi_token_deleted_mail, WebAPIToken),
    ]

    for signal, handler, sender in signal_table:
        signal.connect(handler, sender=sender)


# Fixes bug #3613
_old_header_init = email.header.Header.__init__


def _unified_header_init(self, *args, **kwargs):
    kwargs['continuation_ws'] = b' '

    _old_header_init(self, *args, **kwargs)


email.header.Header.__init__ = _unified_header_init


__all__ = [
    'register_email_hook',
    'unregister_email_hook',
]



from __future__ import unicode_literals

from djblets.testing.decorators import add_fixtures

from reviewboard.notifications.models import WebHookTarget
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class WebHookTargetManagerTests(TestCase):
    """Unit tests for WebHookTargetManager."""

    ENDPOINT_URL = 'http://example.com/endpoint/'

    def test_for_event(self):
        """Testing WebHookTargetManager.for_event"""
        # These should not match.
        WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        WebHookTarget.objects.create(
            events='event3',
            url=self.ENDPOINT_URL,
            enabled=False,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        # These should match.
        target1 = WebHookTarget.objects.create(
            events='event2,event3',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        target2 = WebHookTarget.objects.create(
            events='*',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        targets = WebHookTarget.objects.for_event('event3')
        self.assertEqual(targets, [target1, target2])

    def test_for_event_with_local_site(self):
        """Testing WebHookTargetManager.for_event with Local Sites"""
        site = LocalSite.objects.create(name='test-site')

        # These should not match.
        WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=False,
            local_site=site,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        # This should match.
        target = WebHookTarget.objects.create(
            events='event1,event2',
            url=self.ENDPOINT_URL,
            enabled=True,
            local_site=site,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        targets = WebHookTarget.objects.for_event('event1',
                                                  local_site_id=site.pk)
        self.assertEqual(targets, [target])

    @add_fixtures(['test_scmtools'])
    def test_for_event_with_repository(self):
        """Testing WebHookTargetManager.for_event with repository"""
        repository1 = self.create_repository()
        repository2 = self.create_repository()

        # These should not match.
        unused_target1 = WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=False,
            apply_to=WebHookTarget.APPLY_TO_SELECTED_REPOS)
        unused_target1.repositories.add(repository2)

        unused_target2 = WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=False,
            apply_to=WebHookTarget.APPLY_TO_SELECTED_REPOS)
        unused_target2.repositories.add(repository1)

        WebHookTarget.objects.create(
            events='event3',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_NO_REPOS)

        # These should match.
        target1 = WebHookTarget.objects.create(
            events='event1,event2',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        target2 = WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_SELECTED_REPOS)
        target2.repositories.add(repository1)

        targets = WebHookTarget.objects.for_event('event1',
                                                  repository_id=repository1.pk)
        self.assertEqual(targets, [target1, target2])

    @add_fixtures(['test_scmtools'])
    def test_for_event_with_no_repository(self):
        """Testing WebHookTargetManager.for_event with no repository"""
        repository = self.create_repository()

        # These should not match.
        unused_target1 = WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_SELECTED_REPOS)
        unused_target1.repositories.add(repository)

        WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=False,
            apply_to=WebHookTarget.APPLY_TO_NO_REPOS)

        WebHookTarget.objects.create(
            events='event2',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_NO_REPOS)

        # These should match.
        target1 = WebHookTarget.objects.create(
            events='event1,event2',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_ALL)

        target2 = WebHookTarget.objects.create(
            events='event1',
            url=self.ENDPOINT_URL,
            enabled=True,
            apply_to=WebHookTarget.APPLY_TO_NO_REPOS)

        targets = WebHookTarget.objects.for_event('event1')
        self.assertEqual(targets, [target1, target2])

    def test_for_event_with_all_events(self):
        """Testing WebHookTargetManager.for_event with ALL_EVENTS"""
        with self.assertRaisesMessage(ValueError,
                                      '"*" is not a valid event choice'):
            WebHookTarget.objects.for_event(WebHookTarget.ALL_EVENTS)



from __future__ import unicode_literals

import logging
from xml.dom.minidom import parseString

from django import forms
from django.utils import six
from django.utils.six.moves.urllib.error import HTTPError, URLError
from django.utils.translation import ugettext_lazy as _, ugettext

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceAPIError,
                                            RepositoryError)
from reviewboard.hostingsvcs.forms import (HostingServiceAuthForm,
                                           HostingServiceForm)
from reviewboard.hostingsvcs.service import (HostingService,
                                             HostingServiceClient)
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError


class CodebaseHQAuthForm(HostingServiceAuthForm):
    api_key = forms.CharField(
        label=_('API key'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The API key provided to your Codebase account. This is '
                    'available in My Profile under API Credentials.'))

    domain = forms.CharField(
        label=_('Codebase domain'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The subdomain used to access your Codebase account. '
                    'This is the "<tt>subdomain</tt>" of '
                    '<tt>subdomain</tt>.codebasehq.com.'))

    def get_credentials(self):
        credentials = super(CodebaseHQAuthForm, self).get_credentials()
        credentials.update({
            'domain': self.cleaned_data['domain'],
            'api_key': self.cleaned_data['api_key'],
        })

        return credentials

    class Meta(object):
        help_texts = {
            'hosting_account_username': _(
                'The username you use to log into Codebase. This should '
                '<em>not</em> include the domain name.'
            ),
            'hosting_account_password': _(
                'The password you use to log into Codebase. This is separate '
                'from the API key below.'
            ),
        }


class CodebaseHQForm(HostingServiceForm):
    codebasehq_project_name = forms.CharField(
        label=_('Project name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))

    codebasehq_repo_name = forms.CharField(
        label=_('Repository short name'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The short name of your repository. This can be found by '
                    'clicking the Settings button on the right-hand '
                    'side of the repository browser.'))


class CodebaseHQClient(HostingServiceClient):
    """Client for talking to the Codebase API.

    This implements the API methods that the hosting service needs, converting
    requests into API calls and those back into structured results.
    """

    #: Mimetype used for API requests and responses.
    API_MIMETYPE = 'application/xml'

    def __init__(self, hosting_service):
        """Initialize the client.

        Args:
            hosting_service (CodebaseHQ):
                The hosting service that owns this client.
        """
        self.hosting_service = hosting_service

    def api_get_file(self, repository, project_name, repo_name, path,
                     revision):
        """Return the content of a file in a repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository entry in Review Board.

            project_name (unicode):
                The name of the Codebase project.

            repo_name (unicode):
                The name of the repository.

            path (unicode):
                The path to the file in the repository.

            revision (unicode):
                The revision of the file or commit.

        Returns:
            bytes:
            The contents of the file.
        """
        url = '%s/%s/blob/' % (project_name, repo_name)

        if repository.tool.name == 'Git':
            url += revision
        else:
            if path.startswith('/'):
                path = path[1:]

            url += '%s/%s' % (revision, path)

        return self.api_get(self.build_api_url(url), raw_content=True)

    def api_get_public_keys(self, username):
        """Return information on all public keys for a user.

        Args:
            username (unicode):
                The user to fetch public keys for.

        Returns:
            dict:
            Information on each of the user's public keys.
        """
        return self.api_get(self.build_api_url('users/%s/public_keys'
                                               % username))

    def api_get_repository(self, project_name, repo_name):
        """Return information on a repository.

        Args:
            project_name (unicode):
                The name of the Codebase project.

            repo_name (unicode):
                The name of the repository.

        Returns:
            dict:
            Information on the repository.

            See https://support.codebasehq.com/kb/repositories for the
            data returned.
        """
        return self.api_get(
            self.build_api_url('%s/%s' % (project_name, repo_name)))

    def build_api_url(self, url):
        """Return the URL for an API call.

        Args:
            url (unicode):
                The relative URL for the API call.

        Returns:
            unicode:
            The absolute URL for the API call.
        """
        return 'https://api3.codebasehq.com/%s' % url

    def api_get(self, url, raw_content=False):
        """Perform an HTTP GET request to the API.

        Args:
            url (unicode):
                The full URL to the API resource.

            raw_content (bool, optional):
                If set to ``True``, the raw content of the result will be
                returned, instead of a parsed XML result.

        Returns:
            object:
            The parsed content of the result, as a dictionary, or the raw
            bytes content if ``raw_content`` is ``True``.
        """
        hosting_service = self.hosting_service

        try:
            account_data = hosting_service.account.data
            api_username = '%s/%s' % (account_data['domain'],
                                      hosting_service.account.username)
            api_key = decrypt_password(account_data['api_key'])

            data, headers = self.http_get(
                url,
                username=api_username,
                password=api_key,
                headers={
                    'Accept': self.API_MIMETYPE,
                })

            if raw_content:
                return data
            else:
                return self.parse_xml(data)
        except HTTPError as e:
            data = e.read()
            msg = six.text_type(e)

            rsp = self.parse_xml(data)

            if rsp and 'errors' in rsp:
                errors = rsp['errors']

                if 'error' in errors:
                    msg = errors['error']

            if e.code == 401:
                raise AuthorizationError(msg)
            else:
                raise HostingServiceAPIError(msg, http_code=e.code, rsp=rsp)
        except URLError as e:
            raise HostingServiceAPIError(e.reason)

    def get_xml_text(self, nodes):
        """Return the text contents of a set of XML nodes.

        Args:
            nodes (list of xml.dom.minidom.Element):
                The list of nodes.

        Returns:
            unicode:
            The text content of the nodes.
        """
        return ''.join(
            node.data
            for node in nodes
            if node.nodeType == node.TEXT_NODE
        )

    def parse_xml(self, s):
        """Return the parsed content for an XML document.

        Args:
            s (unicode):
                The XML document as a string.

        Returns:
            dict:
            The parsed content of the XML document, with each key
            being a dictionary of other parsed content.

            If the document cannot be parsed, this will return ``None``.
        """
        try:
            doc = parseString(s)
        except:
            return None

        root = doc.documentElement

        return {
            root.tagName: self._parse_xml_node(root),
        }

    def _parse_xml_node(self, node):
        """Return the parsed content for a node in an XML document.

        This parses the content of a Codebase XML document, turning it into
        arrays, strings, and dictionaries of data.

        Args:
            node (xml.dom.minidom.Element):
                The node being parsed.

        Returns:
            object:
            The parsed content of the node, based on the type of node being
            processed.
        """
        node_type = node.getAttribute('type')
        is_nil = node.getAttribute('nil')

        if node_type == 'array':
            result = [
                self._parse_xml_node(child)
                for child in node.childNodes
                if child.nodeType == child.ELEMENT_NODE
            ]
        elif is_nil == 'true':
            result = None
        else:
            child_nodes = [
                child
                for child in node.childNodes
                if child.nodeType == child.ELEMENT_NODE
            ]

            if child_nodes:
                result = dict([
                    (child.tagName, self._parse_xml_node(child))
                    for child in child_nodes
                ])
            else:
                result = self.get_xml_text(node.childNodes)

        return result


class CodebaseHQ(HostingService):
    """Repository hosting support for Codebase.

    Codebase is a repository hosting service that supports Subversion, Git,
    and Mercurial. It's available at https://codebasehq.com.

    This integration provides repository validation and file fetching. Due to
    API limitations, it does not support post-commit review at this time.
    """

    name = 'Codebase HQ'
    form = CodebaseHQForm
    auth_form = CodebaseHQAuthForm

    needs_authorization = True
    supports_bug_trackers = True
    supports_repositories = True

    supported_scmtools = ['Git', 'Subversion', 'Mercurial']

    repository_fields = {
        'Git': {
            'path': 'git@codebasehq.com:%(domain)s/'
                    '%(codebasehq_project_name)s/'
                    '%(codebasehq_repo_name)s.git',
        },
        'Subversion': {
            'path': 'https://%(domain)s.codebasehq.com/'
                    '%(codebasehq_project_name)s/'
                    '%(codebasehq_repo_name)s.svn',
        },
        'Mercurial': {
            'path': 'https://%(domain)s.codebasehq.com/'
                    'projects/%(codebasehq_project_name)s/repositories/'
                    '%(codebasehq_repo_name)s/',
        },
    }

    bug_tracker_field = (
        'https://%(domain)s.codebasehq.com/projects/'
        '%(codebasehq_project_name)s/tickets/%%s'
    )

    #: A mapping of Codebase SCM types to SCMTool names.
    REPO_SCM_TOOL_MAP = {
        'git': 'Git',
        'svn': 'Subversion',
        'hg': 'Mercurial',
    }

    def __init__(self, *args, **kwargs):
        """Initialize the hosting service.

        Args:
            *args (tuple):
                Positional arguments for the parent constructor.

            **kwargs (dict):
                Keyword arguments for the parent constructor.
        """
        super(CodebaseHQ, self).__init__(*args, **kwargs)

        self.client = CodebaseHQClient(self)

    def authorize(self, username, password, credentials, *args, **kwargs):
        """Authorize an account for Codebase.

        Codebase usees HTTP Basic Auth with an API username (consisting of the
        Codebase team's domain and the account username) and an API key (for
        the password) for API calls, and a standard username/password for
        Subversion repository access. We need to store all of this.

        Args:
            username (unicode):
                The username to authorize.

            password (unicode):
                The API token used as a password.

            credentials (dict):
                Additional credentials from the authentication form.

            *args (tuple):
                Extra unused positional arguments.

            **kwargs (dict):
                Extra unused keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.
        """
        self.account.data.update({
            'domain': credentials['domain'],
            'api_key': encrypt_password(credentials['api_key']),
            'password': encrypt_password(password),
        })

        # Test the account to make sure the credentials are fine. Note that
        # we can only really sanity-check the API token, domain, and username
        # from here. There's no way good way to check the actual password,
        # which we only use for Subversion repositories.
        #
        # This will raise a suitable error message if authorization fails.
        try:
            self.client.api_get_public_keys(username)
        except AuthorizationError:
            raise AuthorizationError(
                ugettext('One or more of the credentials provided were not '
                         'accepted by Codebase.'))

        self.account.save()

    def is_authorized(self):
        """Return if the account has been authorized.

        This checks if all the modern authentication details are stored along
        with the account.

        Returns:
            bool:
            ``True`` if all required credentials are set for the account.
        """
        return (self.account.data.get('api_key') is not None and
                self.account.data.get('password') is not None and
                self.account.data.get('domain') is not None)

    def get_password(self):
        """Return the password for this account.

        This is used primarily for Subversion repositories, so that direct
        access can be performed in order to fetch properties and other
        information.

        This does not return the API key.

        Returns:
            unicode:
            The account password for repository access.
        """
        return decrypt_password(self.account.data['password'])

    def check_repository(self, codebasehq_project_name=None,
                         codebasehq_repo_name=None, tool_name=None,
                         *args, **kwargs):
        """Check the validity of a repository.

        This will perform an API request against Codebase to get information on
        the repository. This will throw an exception if the repository was not
        found, and return cleanly if it was found.

        Args:
            codebase_project_name (unicode):
                The name of the project on Codebase.

            codebasehq_repo_name (unicode):
                The name of the repository on Codebase.

            tool_name (unicode):
                The name of the SCMTool for the repository.

            *args (tuple):
                Extra unused positional arguments passed to this function.

            **kwargs (dict):
                Extra unused keyword arguments passed to this function.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository was not found.
        """
        # The form should enforce these values.
        assert codebasehq_project_name
        assert codebasehq_repo_name
        assert tool_name

        try:
            info = self.client.api_get_repository(codebasehq_project_name,
                                                  codebasehq_repo_name)
        except HostingServiceAPIError as e:
            logging.error('Error finding Codebase repository "%s" for '
                          'project "%s": %s',
                          codebasehq_repo_name, codebasehq_project_name,
                          e)

            raise RepositoryError(
                ugettext('A repository with this name and project was '
                         'not found.'))

        try:
            scm_type = info['repository']['scm']
        except KeyError:
            logging.error('Missing "scm" field for Codebase HQ repository '
                          'payload: %r',
                          info)

            raise RepositoryError(
                ugettext('Unable to determine the type of repository '
                         'from the Codebase API. Please report this.'))

        try:
            expected_tool_name = self.REPO_SCM_TOOL_MAP[scm_type]
        except KeyError:
            logging.error('Unexpected "scm" value "%s" for Codebase HQ '
                          'repository, using payload: %r',
                          scm_type, info)

            raise RepositoryError(
                ugettext('Unable to determine the type of repository '
                         'from the Codebase API. Please report this.'))

        if expected_tool_name != tool_name:
            raise RepositoryError(
                ugettext("The repository type doesn't match what you "
                         "selected. Did you mean %s?")
                % expected_tool_name)

    def get_file(self, repository, path, revision, *args, **kwargs):
        """Returns the content of a file in a repository.

        This will perform an API request to fetch the contents of a file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository containing the file.

            path (unicode):
                The path to the file in the repository.

            revision (unicode):
                The revision of the file in the repository.

            *args (tuple):
                Extra unused positional arguments passed to this function.

            **kwargs (dict):
                Extra unused keyword arguments passed to this function.

        Returns:
            byets:
            The content of the file in the repository.
        """
        try:
            return self.client.api_get_file(
                repository,
                repository.extra_data['codebasehq_project_name'],
                repository.extra_data['codebasehq_repo_name'],
                path, revision)
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise FileNotFoundError(path, revision)
            else:
                logging.warning('Failed to fetch file from Codebase HQ '
                                'repository %s: %s',
                                repository, e)
                raise

    def get_file_exists(self, repository, path, revision, *args, **kwargs):
        """Returns whether a given file exists.

        This will perform an API request to fetch the contents of a file,
        returning ``True`` if the content could be fetched.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository containing the file.

            path (unicode):
                The path to the file in the repository.

            revision (unicode):
                The revision of the file in the repository.

            *args (tuple):
                Extra unused positional arguments passed to this function.

            **kwargs (dict):
                Extra unused keyword arguments passed to this function.

        Returns:
            bool:
            ``True`` if the file exists in the repository.
        """
        try:
            self.client.api_get_file(
                repository,
                repository.extra_data['codebasehq_project_name'],
                repository.extra_data['codebasehq_repo_name'],
                path, revision)

            return True
        except HostingServiceAPIError:
            return False



"""Unit tests for reviewboard.accounts.models.ReviewRequestVisit."""

from __future__ import unicode_literals

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.testing import TestCase


class ReviewRequestVisitTests(TestCase):
    """Unit tests for reviewboard.accounts.models.ReviewRequestVisit."""

    fixtures = ['test_users']

    def test_default_visibility(self):
        """Testing ReviewRequestVisit.visibility default value"""
        review_request = self.create_review_request(publish=True)
        self.client.login(username='admin', password='admin')
        self.client.get(review_request.get_absolute_url())

        visit = ReviewRequestVisit.objects.get(
            user__username='admin', review_request=review_request.id)

        self.assertEqual(visit.visibility, ReviewRequestVisit.VISIBLE)
